import sys
try:
	import json
except ImportError:
	import simplejson as json

from collections import OrderedDict


def dataToPrettyJson(data, ensure_ascii=False, sort_keys=False):
	return json.dumps(
		data,
		sort_keys=sort_keys,
		indent=2,
		ensure_ascii=ensure_ascii,
	)


def dataToCompactJson(data, ensure_ascii=False, sort_keys=False):
	return json.dumps(
		data,
		sort_keys=sort_keys,
		separators=(',', ':'),
		ensure_ascii=ensure_ascii,
	)


jsonToData = json.loads


def jsonToOrderedData(text):
	return json.JSONDecoder(
		object_pairs_hook=OrderedDict,
	).decode(text)

###############################


def loadJsonConf(module, confPath, decoders={}):
	from os.path import isfile
	###
	if not isfile(confPath):
		return
	###
	try:
		text = open(confPath).read()
	except Exception as e:
		print('failed to read file "%s": %s' % (confPath, e))
		return
	###
	try:
		data = json.loads(text)
	except Exception as e:
		print('invalid json file "%s": %s' % (confPath, e))
		return
	###
	if isinstance(module, str):
		module = sys.modules[module]
	for param, value in data.items():
		try:
			decoder = decoders[param]
		except KeyError:
			pass
		else:
			value = decoder(value)
		setattr(module, param, value)


def saveJsonConf(module, confPath, params, encoders={}):
	if isinstance(module, str):
		module = sys.modules[module]
	###
	data = OrderedDict()
	for param in params:
		value = getattr(module, param)
		try:
			encoder = encoders[param]
		except KeyError:
			pass
		else:
			value = encoder(value)
		data[param] = value
	###
	text = dataToPrettyJson(data)
	try:
		open(confPath, 'w').write(text)
	except Exception as e:
		print('failed to save file "%s": %s' % (confPath, e))
		return


def loadModuleJsonConf(module):
	if isinstance(module, str):
		module = sys.modules[module]
	###
	decoders = getattr(module, 'confDecoders', {})
	###
	try:
		sysConfPath = module.sysConfPath
	except AttributeError:
		pass
	else:
		loadJsonConf(
			module,
			sysConfPath,
			decoders,
		)
	####
	loadJsonConf(
		module,
		module.confPath,
		decoders,
	)
	# should use module.confParams to restrict json keys? FIXME


def saveModuleJsonConf(module):
	if isinstance(module, str):
		module = sys.modules[module]
	###
	saveJsonConf(
		module,
		module.confPath,
		module.confParams,
		getattr(module, 'confEncoders', {}),
	)



# -*- coding: utf-8 -*-
# appledict/_dict.py
# Output to Apple Dictionary xml sources for Dictionary Development Kit.
#
# Copyright (C) 2016 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Copyright (C) 2016 Ratijas <ratijas.t@me.com>
# Copyright (C) 2012-2015 Xiaoqiang Wang <xiaoqiangwang AT gmail DOT com>
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

import logging
import re
import string
from xml.sax.saxutils import unescape, quoteattr



from . import _normalize
from pyglossary.plugins.formats_common import log, toStr

log = logging.getLogger("root")



def get_beautiful_soup():
	try:
		import bs4 as BeautifulSoup
	except ImportError:
		try:
			import BeautifulSoup
		except ImportError:
			return None
	if int(BeautifulSoup.__version__.split(".")[0]) < 4:
		raise ImportError(
			"BeautifulSoup is too old, required at least version 4, " +
			"%r found.\n" % BeautifulSoup.__version__ +
			"Please run `sudo pip3 install lxml beautifulsoup4 html5lib`"
		)
	return BeautifulSoup


digs = string.digits + string.ascii_letters


def base36(x):
	"""
	simplified version of int2base
	http://stackoverflow.com/questions/2267362/convert-integer-to-a-string-in-a-given-numeric-base-in-python#2267446
	"""
	digits = []
	while x:
		digits.append(digs[x % 36])
		x //= 36
	digits.reverse()
	return "".join(digits)


def id_generator():
	cnt = 1

	while True:
		s = "_%s" % base36(cnt)
		yield s
		cnt += 1


def indexes_generator(indexes_lang):
	"""
	factory that acts according to glossary language

	:param indexes_lang: str
	"""
	indexer = None
	"""Callable[[Sequence[str], str], Sequence[str]]"""
	if indexes_lang:
		from . import indexes as idxs
		indexer = idxs.languages.get(indexes_lang, None)
		if not indexer:
			msg = "extended indexes not supported for the specified language: %s.\n"\
				  "following languages avaible: %s." %\
				  (indexes_lang, ", ".join(list(idxs.languages.keys())))
			log.error(msg)
			raise ValueError(msg)

	def generate_indexes(title, alts, content, BeautifulSoup):
		indexes = [title]
		indexes.extend(alts)

		if BeautifulSoup:
			quoted_title = BeautifulSoup.dammit.EntitySubstitution.substitute_xml(title, True)
		else:
			quoted_title = '"%s"' % title.replace(">", "&gt;").replace('"', "&quot;")

		if indexer:
			indexes = set(indexer(indexes, content))

		normal_indexes = set()
		for idx in indexes:
			normal = _normalize.title(idx, BeautifulSoup)
			normal_indexes.add(_normalize.title_long(normal))
			normal_indexes.add(_normalize.title_short(normal))
		normal_indexes.discard(title)

		normal_indexes = [s for s in normal_indexes if s.strip()]
		# skip empty titles.  everything could happen.

		s = "<d:index d:value=%s d:title=%s/>" % (quoted_title, quoted_title)
		if BeautifulSoup:
			for idx in normal_indexes:
				s += "<d:index d:value=%s d:title=%s/>" % (
					BeautifulSoup.dammit.EntitySubstitution.substitute_xml(idx, True),
					quoted_title)
		else:
			for idx in normal_indexes:
				s += '<d:index d:value="%s" d:title=%s/>' % (
					idx.replace(">", "&gt;").replace('"', "&quot;"),
					quoted_title)
		return s
	return generate_indexes


close_tag = re.compile("<(BR|HR)>", re.IGNORECASE)
nonprintable = re.compile("[\x00-\x07\x0e-\x1f]")
img_tag = re.compile("<IMG (.*?)>", re.IGNORECASE)

em0_9_re = re.compile(r'<div style="margin-left:(\d)em">')
em0_9_sub = r'<div class="m\1">'

em0_9_ex_re = re.compile(r'<div class="ex" style="margin-left:(\d)em;color:steelblue">')
em0_9_ex_sub = r'<div class="m\1 ex">'

href_re = re.compile(r'''href=(["'])(.*?)\1''')


def href_sub(x):
	href = x.groups()[1]
	if href.startswith("http"):
		return x.group()
	if href.startswith("bword://"):
		href = href[len("bword://"):]
	return "href=%s" % quoteattr(
		"x-dictionary:d:" + unescape(
			href,
			{"&quot;": '"'},
		)
	)


def is_green(x):
	return "color:green" in x.get("style", "")

margin_re = re.compile("margin-left:(\d)em")


def remove_style(tag, line):
	s = "".join(tag["style"].replace(line, "").split(";"))
	if s:
		tag["style"] = s
	else:
		del tag["style"]


def format_clean_content(title, body, BeautifulSoup):
	# heavily integrated with output of dsl reader plugin!
	# and with xdxf also.
	"""
	:param title: str | None
	"""

	# class="sec" => d:priority="2"
	# style="color:steelblue" => class="ex"
	# class="p" style="color:green" => class="p"
	# style="color:green" => class="c"
	# style="margin-left:{}em" => class="m{}"
	# <s> => <del>

	# xhtml is strict
	if BeautifulSoup:
		soup = BeautifulSoup.BeautifulSoup(body, "lxml", from_encoding="utf-8")
		# difference between "lxml" and "html.parser"
		if soup.body:
			soup = soup.body

		for tag in soup(class_="sec"):
			tag["class"].remove("sec")
			if not tag["class"]:
				del tag["class"]
			tag["d:priority"] = "2"
		for tag in soup(lambda x: "color:steelblue" in x.get("style", "")):
			remove_style(tag, "color:steelblue")
			if "ex" not in tag.get("class", []):
				tag["class"] = tag.get("class", []) + ["ex"]
		for tag in soup(is_green):
			remove_style(tag, "color:green")
			if "p" not in tag.get("class", ""):
				tag["class"] = tag.get("class", []) + ["c"]
		for tag in soup(True):
			if "style" in tag.attrs:
				m = margin_re.search(tag["style"])
				if m:
					remove_style(tag, m.group(0))
					tag["class"] = tag.get("class", []) + ["m" + m.group(1)]
		for tag in soup.select("[href]"):
			href = tag["href"]
			if href.startswith("bword://"):
				href = href[len("bword://"):]
			if not (href.startswith("http:") or href.startswith("https:")):
				tag["href"] = "x-dictionary:d:%s" % href
		for tag in soup("u"):
			tag.name = "span"
			tag["class"] = tag.get("class", []) + ["u"]
		for tag in soup("s"):
			tag.name = "del"

		if title:
			h1 = BeautifulSoup.Tag(name="h1")
			h1.string = title
			soup.insert(0, h1)
		# hence the name BeautifulSoup
		content = toStr(soup.encode_contents())
	else:
		# somewhat analogue to what BeautifulSoup suppose to do
		body = em0_9_re.sub(em0_9_sub, body)
		body = em0_9_ex_re.sub(em0_9_ex_sub, body)
		body = href_re.sub(href_sub, body)

		body = body \
			.replace('<i style="color:green">', '<i class="c">') \
			.replace('<i class="p" style="color:green">', '<i class="p">') \
			.replace('<span class="ex" style="color:steelblue">', '<span class="ex">') \
			.replace('<span class="sec ex" style="color:steelblue">', '<span class="sec ex">') \
			.replace('<u>', '<span class="u">').replace('</u>', '</span>') \
			.replace('<s>', '<del>').replace('</s>', '</del>')

		# nice header to display
		content = "<h1>%s</h1>%s" % (title, body) if title else body
		content = close_tag.sub("<\g<1> />", content)
		content = img_tag.sub("<img \g<1>/>", content)
	content = content.replace("&nbsp;", "&#160;")
	content = nonprintable.sub("", content)
	return content





# -*- coding: utf-8 -*-

from heapq import heappush, heappop
from heapq import merge

import logging
log = logging.getLogger('root')


def hsortStream(stream, maxHeapSize, key=None):
	"""
		stream: a generator or iterable
		maxHeapSize: int, maximum size of heap
		key: a key function, as in `list.sort` method, or `sorted` function
			 if key is None, we consume less memory

		the sort is Stable (unlike normal heapsort) because we include the
			index (after item / output of key function)
	"""
	hp = []
	if key:
		for index, item in enumerate(stream):
			if len(hp) >= maxHeapSize:
				yield heappop(hp)[2]
			heappush(hp, (
				key(item),  # for sorting order
				index,  # for sort being Stable
				item,  # for fetching result
			))
		while hp:
			yield heappop(hp)[2]
	else:  # consume less memory
		for index, item in enumerate(stream):
			if len(hp) >= maxHeapSize:
				yield heappop(hp)[0]
			heappush(hp, (
				item,  # for sorting order, and fetching result
				index,  # for sort being Stable
			))
		while hp:
			yield heappop(hp)[0]


def hsortStreamList(streams, *args, **kwargs):
	streams = [
		 hsortStream(stream, *args, **kwargs)
		 for stream in streams
	]
	return merge(*tuple(streams))


def stdinIntegerStream():
	while True:
		line = input(' Input item: ')
		if not line:
			break
		yield int(line)


def stdinStringStream():
	while True:
		line = raw_input(' Input item: ')
		if not line:
			break
		yield line


def randomChoiceGenerator(choices, count):
	import random
	for _ in range(count):
		yield random.choice(choices)


def test_hsortStreamList(count=10):
	for item in hsortStreamList(
		[
			randomChoiceGenerator(range(0, 50), count),
			randomChoiceGenerator(range(30, 50), count),
			randomChoiceGenerator(range(10, 40), count),
		],
		maxHeapSize=5,
		key=None,
	):
		print(item)


def main():
	test_hsortStreamList()
#	stream = stdinIntegerStream()
#	for line in hsortStream(stream, 3):
#		print('------ Placed item: %s'%line)

if __name__ == '__main__':
	main()



from itertools import (
	takewhile,
	repeat,
)


def toBytes(s):
	return bytes(s, 'utf8') if isinstance(s, str) else bytes(s)


def fileCountLines(filename, newline='\n'):
	newline = toBytes(newline)  # required? FIXME
	f = open(filename, 'rb')  # or 'r'
	bufgen = takewhile(
		lambda x: x, (f.read(1024*1024) for _ in repeat(None))
	)
	return sum(
		buf.count(newline) for buf in bufgen if buf
	)


class FileLineWrapper(object):
	def __init__(self, f):
		self.f = f
		self.line = 0

	def close(self):
		return self.f.close()

	def readline(self):
		self.line += 1
		return self.f.readline()

	def __iter__(self):
		return iter(self.f)



# -*- coding: utf-8 -*-

from formats_common import *
import subprocess

enable = True
format = "Treedict"
description = "TreeDict"
extentions = [".tree", ".treedict"]
readOptions = []
writeOptions = [
	"encoding",  # str
]


def write(glos, filename, encoding="utf-8", archive="tar.bz2", sep=os.sep):
	if os.path.exists(filename):
		if os.path.isdir(filename):
			if os.listdir(filename):
				log.warning("Warning: directory \"%s\" is not empty.")
		else:
			raise IOError("\"%s\" is not a directory")
	for entry in glos:
		defi = entry.getDefi()
		for word in entry.getWords():
			if not word:
				log.error("empty word")
				continue
			chars = list(word)
			try:
				os.makedirs(filename + os.sep + sep.join(chars[:-1]))
			except:
				pass
			entryFname = "%s%s%s.m" % (
				filename,
				os.sep,
				sep.join(chars),
			)
			try:
				with open(entryFname, "a", encoding=encoding) as entryFp:
					entryFp.write(defi)
			except:
				log.exception("")
	if archive:
		if archive == "tar.gz":
			(output, error) = subprocess.Popen(
				["tar", "-czf", filename+".tar.gz", filename],
				stdout=subprocess.PIPE
			).communicate()
		elif archive == "tar.bz2":
			(output, error) = subprocess.Popen(
				["tar", "-cjf", filename+".tar.bz2", filename],
				stdout=subprocess.PIPE
			).communicate()
		elif archive == "zip":
			(output, error) = subprocess.Popen(
				["zip", "-r", filename+".zip", filename],
				stdout=subprocess.PIPE
			).communicate()
		else:
			log.error("Undefined archive format: \"%s\"" % archive)
		try:
			shutil.rmtree(filename, ignore_errors=True)
		except:
			pass



import logging
import traceback
import inspect
from pprint import pformat
import sys
import os
from os.path import (
	join,
	isfile,
	isdir,
	exists,
	realpath,
	dirname,
)
import platform

VERSION = "3.2.1"

class MyLogger(logging.Logger):
	levelsByVerbosity = (
		logging.CRITICAL,
		logging.ERROR,
		logging.WARNING,
		logging.INFO,
		logging.DEBUG,
		logging.NOTSET,
	)
	levelNamesCap = [
		"Critical",
		"Error",
		"Warning",
		"Info",
		"Debug",
		"All",  # "Not-Set",
	]

	def setVerbosity(self, verbosity):
		self.setLevel(self.levelsByVerbosity[verbosity])
		self._verbosity = verbosity

	def getVerbosity(self):
		return getattr(self, "_verbosity", 3)  # FIXME

	def pretty(self, data, header=""):
		self.debug(header + pformat(data))

	def isDebug(self):
		return self.getVerbosity() >= 4

def format_var_dict(dct, indent=4, max_width=80):
	lines = []
	pre = " " * indent
	for key, value in dct.items():
		line = pre + key + " = " + repr(value)
		if len(line) > max_width:
			line = line[:max_width-3] + "..."
			try:
				value_len = len(value)
			except:
				pass
			else:
				line += "\n" + pre + "len(%s) = %s"%(key, value_len)
		lines.append(line)
	return "\n".join(lines)


def format_exception(exc_info=None, add_locals=False, add_globals=False):
	if not exc_info:
		exc_info = sys.exc_info()
	_type, value, tback = exc_info
	text = "".join(traceback.format_exception(_type, value, tback))

	if add_locals or add_globals:
		try:
			frame = inspect.getinnerframes(tback, context=0)[-1][0]
		except IndexError:
			pass
		else:
			if add_locals:
				text += "Traceback locals:\n%s\n" % format_var_dict(
					frame.f_locals,
				)
			if add_globals:
				text += "Traceback globals:\n%s\n" % format_var_dict(
					frame.f_globals,
				)

	return text


class StdLogHandler(logging.Handler):
	startRed = "\x1b[31m"
	endFormat = "\x1b[0;0;0m"  # len=8

	def __init__(self, noColor=False):
		logging.Handler.__init__(self)
		self.noColor = noColor

	def emit(self, record):
		msg = record.getMessage()
		###
		if record.exc_info:
			_type, value, tback = record.exc_info
			tback_text = format_exception(
				exc_info=record.exc_info,
				add_locals=(log.level <= logging.DEBUG),  # FIXME
				add_globals=False,
			)

			if not msg:
				msg = "unhandled exception:"
			msg += "\n"
			msg += tback_text
		###
		if record.levelname in ("CRITICAL", "ERROR"):
			if not self.noColor:
				msg = self.startRed + msg + self.endFormat
			fp = sys.stderr
		else:
			fp = sys.stdout
		###
		fp.write(msg + "\n")
		fp.flush()
#	def exception(self, msg):
#		if not self.noColor:
#			msg = self.startRed + msg + self.endFormat
#		sys.stderr.write(msg + "\n")
#		sys.stderr.flush()


def checkCreateConfDir():
	if not isdir(confDir):
		if exists(confDir):  # file, or anything other than directory
			os.rename(confDir, confDir + ".bak")  # we do not import old config
		os.mkdir(confDir)
	if not exists(userPluginsDir):
		os.mkdir(userPluginsDir)
	if not isfile(confJsonFile):
		with open(rootConfJsonFile) as srcFp, \
		  open(confJsonFile, "w") as userFp:
			userFp.write(srcFp.read())


# __________________________________________________________________________ #

logging.setLoggerClass(MyLogger)
log = logging.getLogger("root")

sys.excepthook = lambda *exc_info: log.critical(
	format_exception(
		exc_info=exc_info,
		add_locals=(log.level <= logging.DEBUG),  # FIXME
		add_globals=False,
	)
)

sysName = platform.system()

# can set env var WARNINGS to: "error", "ignore", "always", "default", "module", "once"
if os.getenv("WARNINGS"):
	import warnings
	warnings.filterwarnings(os.getenv("WARNINGS"))


if hasattr(sys, "frozen"):
	rootDir = dirname(sys.executable)
	uiDir = join(rootDir, "ui")
else:
	uiDir = dirname(realpath(__file__))
	rootDir = dirname(uiDir)

dataDir = rootDir
if dataDir.endswith("dist-packages") or dataDir.endswith("site-packages"):
	dataDir = dirname(sys.argv[0])

appResDir = join(dataDir, "res")

if os.sep == "/":  # Operating system is Unix-Like
	homeDir = os.getenv("HOME")
	user = os.getenv("USER")
	tmpDir = "/tmp"
	# os.name == "posix" # FIXME
	if sysName == "Darwin":  # MacOS X
		confDir = homeDir + "/Library/Preferences/PyGlossary"
		# or maybe: homeDir + "/Library/PyGlossary"
		# os.environ["OSTYPE"] == "darwin10.0"
		# os.environ["MACHTYPE"] == "x86_64-apple-darwin10.0"
		# platform.dist() == ("", "", "")
		# platform.release() == "10.3.0"
	else:  # GNU/Linux, ...
		confDir = homeDir + "/.pyglossary"
elif os.sep == "\\":  # Operating system is Windows
	homeDir = os.getenv("HOMEDRIVE") + os.getenv("HOMEPATH")
	user = os.getenv("USERNAME")
	tmpDir = os.getenv("TEMP")
	confDir = os.getenv("APPDATA") + "\\" + "PyGlossary"
else:
	raise RuntimeError(
		"Unknown path seperator(os.sep==%r)" % os.sep +
		", unknown operating system!"
	)

confJsonFile = join(confDir, "config.json")
rootConfJsonFile = join(dataDir, "config.json")
userPluginsDir = join(confDir, "plugins")



# -*- coding: utf-8 -*-

from formats_common import *

enable = True
format = "Freedict"
description = "FreeDict (tei)"
extentions = [".tei"]
readOptions = []


def write(glos, filename):
	fp = open(filename, "w")

	fp.write("""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE TEI.2 PUBLIC "-//TEI P3//DTD Main Document Type//EN"
"/usr/share/sgml/tei-3/tei2.dtd" [
<!ENTITY %% TEI.dictionaries "INCLUDE" > ]>
<tei.2>
<teiHeader>
<fileDesc>
<titleStmt>
	<title>%s</title>
	<respStmt><resp>converted with</resp><name>PyGlossary</name></respStmt>
</titleStmt>
<publicationStmt><p>freedict.de</p></publicationStmt>
<sourceDesc><p>%s</p></sourceDesc>
</fileDesc>
</teiHeader>
<text><body>""" % (glos.getInfo("title"), filename))

	for entry in glos:
		if entry.isData():
			# FIXME
			continue
		word = entry.getWord()
		defi = entry.getDefi()
		fp.write("""<entry>
<form><orth>%s</orth></form>
<gramgrp><pos>n</pos></gramgrp>
<trans><tr>%s</tr></trans>
</entry>""" % (word, defi))
	fp.write("</body></text></tei.2>")
	fp.close()



# -*- coding: utf-8 -*-
# http://www.octopus-studio.com/download.en.htm

from formats_common import *
from collections import OrderedDict

enable = True
format = "OctopusMdictSource"
description = "Octopus MDict Source"
extentions = [".mtxt"]
readOptions = [
	"encoding",  # str
	"links", # bool, support for inconsecutive links, #68
]
writeOptions = [
	"resources",  # bool
]


def read(
	glos,
	filename,
	encoding="utf-8",
	links=False,
):
	with open(filename, encoding=encoding) as fp:
		text = fp.read()
	text = text.replace("\r\n", "\n")
	text = text.replace("entry://", "bword://")
	lastEntry = None
	linksDict = {}
	entryDict = OrderedDict()
	for section in text.split("</>"):
		lines = section.strip().split("\n")
		if len(lines) < 2:
			continue
		word = lines[0]
		defi = "\n".join(lines[1:])

		if defi.startswith("@@@LINK="):
			mainWord = defi.partition("=")[2]
			if links:
				linksDict[word] = mainWord
			elif lastEntry and lastEntry.getWords()[0] == mainWord:
				lastEntry.addAlt(word)
			else:
				log.error("alternate is not ride after word: %s" % defi)
			continue

		entry = glos.newEntry(word, defi)
		if links:
			# do not call glos.addEntry or glos.addEntryObj
			# because we will need to modify entries at the end
			# and we need to keep a OrderedDict of entries (ordered list of entries, and dict of entries by word)
			entryDict[word] = entry
		else:
			if lastEntry:
				# now that we know there are no more alternate forms of lastEntry
				glos.addEntryObj(lastEntry)
			lastEntry = entry

	if links:
		for sourceWord, targetWord in linksDict.items():
			targetEntry = entryDict.get(targetWord)
			if targetEntry is None:
				log.error("Link to non-existing word %s" % targetWord)
				continue
			targetEntry.addAlt(sourceWord)
		for entry in entryDict.values():
			glos.addEntryObj(entry)
	else:
		if lastEntry:
			glos.addEntryObj(lastEntry)

def writeEntryGen(glos):
	for entry in glos:
		words = entry.getWords()
		defis = entry.getDefis()

		yield glos.newEntry(words[0], defis)

		for alt in words[1:]:
			yield glos.newEntry(
				alt,
				"@@@LINK=%s" % words[0],
			)


def write(
	glos,
	filename,
	resources=True,
):
	glos.writeTxt(
		"\n",
		"\n</>\n",
		filename=filename,
		writeInfo=False,
		rplList=[
			("bword://", "entry://"),
		],
		ext=".mtxt",
		head="",
		iterEntries=writeEntryGen(glos),
		newline="\r\n",
		resources=resources,
	)




from sqlalchemy import *
from elixir import *


class Entry(Entity):
	has_field('s_id', Integer)
	has_field('wname', Unicode)
	has_field('wmean', Unicode)


setup_all()


objectstore = SessionContext()


def writeSqlite_ex(glos, filename=''):
	metadata = MetaData()
	metadata.bind = 'sqlite:///' + filename
	metadata.create_all()
	d = glos.data
	n = len(d)
	for i in range(n):
		Entry(s_id=i+1, wname=d[i][0], wmean=d[i][1])
	objectstore.flush()
	# objectstore.clear()



def takePhonetic_oxford_gb(glos):
	phonGlos = Glossary() ## phonetic glossary
	phonGlos.setInfo('name', glos.getInfo('name') + '_phonetic')
	for entry in glos:
		word = entry.getWord()
		defi = entry.getDefi()
		if not defi.startswith('/'):
			continue
		#### Now set the phonetic to the `ph` variable.
		ph = ''
		for s in (
			'/ adj',
			'/ v',
			'/ n',
			'/ adv',
			'/adj',
			'/v',
			'/n',
			'/adv',
			'/ n',
			'/ the',
		):
			i = defi.find(s, 2, 85)
			if i==-1:
				continue
			else:
				ph = defi[:i+1]
				break
		ph = ph.replace(';', '\t')\
			   .replace(',', '\t')\
			   .replace('     ', '\t')\
			   .replace('    ', '\t')\
			   .replace('  ', '\t')\
			   .replace('//', '/')\
			   .replace('\t/\t', '\t')\
			   .replace('<i>US</i>\t', '\tUS: ')\
			   .replace('<i>US</i>', '\tUS: ')\
			   .replace('\t\t\t', '\t')\
			   .replace('\t\t', '\t')\
		#	  .replace('/', '')
		#	  .replace('\\n ', '\\n')
		#	  .replace('\\n ', '\\n')
		if ph != '':
			phonGlos.addEntry(word, ph)
	return phonGlos




# -*- coding: utf-8 -*-
# progressbar.py - Text progressbar library for python
#
# Copyright (C) 2006 Nilton Volpato <nilton.volpato@gmail.com>
#					(until version 2.2)
# Copyright (C) 2009-2010 Saeed Rasooli <saeed.gnu@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. Or on Debian systems, from:
# /usr/share/common-licenses/LGPL
# If not, see <http://www.gnu.org/licenses/lgpl.txt>.


"""
	Text progressbar library for python.

	This library provides a text mode progressbar. This is tipically used
	to display the progress of a long running operation, providing a
	visual clue that processing is underway.

	The ProgressBar class manages the progress, and the format of the line
	is given by a number of widgets. A widget is an object that may
	display diferently depending on the state of the progress. There are
	three types of widget:
	- a string, which always shows itself;
	- a ProgressBarWidget, which may return a diferent value every time
	it's update method is called; and
	- a ProgressBarWidgetHFill, which is like ProgressBarWidget, except it
	expands to fill the remaining width of the line.

	The progressbar module is very easy to use, yet very powerful. And
	automatically supports features like auto-resizing when available.
"""

# Changelog
#
# 2009-??-??: used in PyGlossary by Saeed Rasooli with some modifications
# 2006-05-07: v2.2 fixed bug in windows
# 2005-12-04: v2.1 autodetect terminal width, added start method
# 2005-12-04: v2.0 everything is now a widget (wow!)
# 2005-12-03: v1.0 rewrite using widgets
# 2005-06-02: v0.5 rewrite
# 2004-??-??: v0.1 first version

import sys
import time
from array import array
import signal

import logging
log = logging.getLogger('root')


class ProgressBarWidget(object):
	"""
		This is an element of ProgressBar formatting.

		The ProgressBar object will call it's update value when an update
		is needed. It's size may change between call, but the results will
		not be good if the size changes drastically and repeatedly.
	"""
	def update(self):
		"""
			Returns the string representing the widget.

			The parameter pbar is a reference to the calling ProgressBar,
			where one can access attributes of the class for knowing how
			the update must be made.

			At least this function must be overriden.
		"""
		pass

	def __str__(self):
		return self.update()

	def __add__(self, other):
		if isinstance(other, bytes):
			return str(self) + other.encode('utf-8')
		return str(self) + str(other)

	def __radd__(self, other):
		if isinstance(other, bytes):
			return other.encode('utf-8') + str(self)
		return str(other) + str(self)


class ProgressBarWidgetHFill(object):
	"""
		This is a variable width element of ProgressBar formatting.

		The ProgressBar object will call it's update value, informing the
		width this object must the made. This is like TeX \\hfill, it will
		expand to fill the line. You can use more than one in the same
		line, and they will all have the same width, and together will
		fill the line.
	"""
	def update(self, width):
		"""
			Returns the string representing the widget.

			The parameter pbar is a reference to the calling ProgressBar,
			where one can access attributes of the class for knowing how
			the update must be made. The parameter width is the total
			horizontal width the widget must have.

			At least this function must be overriden.
		"""
		pass


class ETA(ProgressBarWidget):
	"""
		Widget for the Estimated Time of Arrival
	"""
	def __init__(self, text='ETA: '):
		self.text = text

	def format_time(self, seconds):
		return time.strftime('%H:%M:%S', time.gmtime(seconds))

	def update(self):
		pbar = self.pbar
		if pbar.currval == 0:
			return 'ETA: --:--:--'
		elif pbar.finished:
			return 'Time: %s' % self.format_time(pbar.seconds_elapsed)
		else:
			elapsed = pbar.seconds_elapsed
			eta = elapsed * pbar.maxval / pbar.currval - elapsed
			return '%s%s' % (self.text, self.format_time(eta))


class FileTransferSpeed(ProgressBarWidget):
	"""
		Widget for showing the transfer speed (useful for file transfers).
	"""
	def __init__(self):
		self.fmt = '%6.2f %s'
		self.units = ['B', 'K', 'M', 'G', 'T', 'P']

	def update(self):
		pbar = self.pbar
		if pbar.seconds_elapsed < 2e-6:  # == 0:
			bps = 0.0
		else:
			bps = float(pbar.currval) / pbar.seconds_elapsed
		spd = bps
		for u in self.units:
			if spd < 1000:
				break
			spd /= 1000
		return self.fmt % (spd, u+'/s')


class RotatingMarker(ProgressBarWidget):
	"""
		A rotating marker for filling the bar of progress.
	"""
	def __init__(self, markers='|/-\\'):
		# Some cool exmaple for markers:
		# u'░▒▓█'
		# u'⬅⬉⬆⬈➡⬊⬇⬋' , u'⬌⬉⬆⬈⬌⬊⬇⬋', u'➚➙➘➙', u'➝➞➡➞'
		# u' ⚊⚌☰⚌⚊', u' ⚋⚊⚍⚌☱☰☱⚌⚍⚊⚋' ,
		# '<(|)>)|(', u'❘❙❚❙', u'❢❣❤❣'
		self.markers = markers
		self.curmark = -1

	def __len__(self):
		return 1

	def update(self):
		pbar = self.pbar
		if pbar.finished:
			return self.markers[0]
		self.curmark = (self.curmark + 1) % len(self.markers)
		return self.markers[self.curmark]


class Percentage(ProgressBarWidget):
	"""
		Just the percentage done.
	"""
	def update(self):
		pbar = self.pbar
		return '%5.1f' % pbar.percentage()


class Bar(ProgressBarWidgetHFill):
	"""
		The bar of progress. It will strech to fill the line.
	"""
	def __init__(self, marker='#', left='|', right='|'):
		self.marker = marker
		self.left = left
		self.right = right

	def _format_marker(self, pbar):
		if type(self.marker) in (str, str):
			return self.marker
		else:
			return self.marker.update(pbar)

	def update(self, width):
		width = int(width)
		pbar = self.pbar
		percent = pbar.percentage()
		cwidth = width - len(self.left) - len(self.right)
		marked_width = int(percent * cwidth / 100.0)
		m = self._format_marker(pbar)
		bar = (self.left + (m*marked_width).ljust(cwidth) + self.right)
		return bar


class ReverseBar(Bar):
	"""
		The reverse bar of progress, or bar of regress. :)
	"""
	def update(self, width):
		pbar = self.pbar
		percent = pbar.percentage()
		cwidth = width - len(self.left) - len(self.right)
		marked_width = int(percent * cwidth / 100.0)
		m = self._format_marker(pbar)
		bar = (self.left + (m*marked_width).rjust(cwidth) + self.right)
		return bar


class ProgressBar(object):
	"""
		This is the ProgressBar class, it updates and prints the bar.

		The term_width parameter may be an integer. Or None, in which case
		it will try to guess it, if it fails it will default to 80 columns.

		The simple use is like this:
		>>> pbar = ProgressBar().start()
		>>> for i in xrange(100):
		...    # do something
		...    pbar.update(i+1)
		...
		>>> pbar.finish()

		But anything you want to do is possible (well, almost anything).
		You can supply different widgets of any type in any order. And you
		can even write your own widgets! There are many widgets already
		shipped and you should experiment with them.

		When implementing a widget update method you may access any
		attribute or function of the ProgressBar object calling the
		widget's update method. The most important attributes you would
		like to access are:
		- currval: current value of the progress, 0 <= currval <= maxval
		- maxval: maximum (and final) value of the progress
		- finished: True if the bar is have finished (reached 100%), False o/w
		- start_time: first time update() method of ProgressBar was called
		- seconds_elapsed: seconds elapsed since start_time
		- percentage(): percentage of the progress (this is a method)
	"""
	default_widgets = [Percentage(), ' ', Bar()]

	def __init__(
		self,
		maxval=100.0,
		widgets=None,
		update_step=0.1,
		term_width=None,
		fd=sys.stderr,
	):
		assert maxval > 0
		self.maxval = maxval

		if widgets is None:
			widgets = self.default_widgets
		self.widgets = widgets

		for w in self.widgets:
			# log.debug( type(w) is ProgressBarWidget )
			# if not isinstance(w, str):
			try:
				w.pbar = self
			except:
				pass
		self.update_step = update_step
		self.fd = fd
		self.signal_set = False
		if term_width is None:
			try:
				self.handle_resize(None, None)
				signal.signal(signal.SIGWINCH, self.handle_resize)
				self.signal_set = True
			except:
				self.term_width = 79
		else:
			self.term_width = term_width

		self.currval = 0
		self.finished = False
		self.prev_percentage = -1
		self.start_time = None
		self.seconds_elapsed = 0

	def handle_resize(self, signum, frame):
		try:
			from fcntl import ioctl
			import termios
		except:
			pass
		else:
			h, w = array('h', ioctl(self.fd, termios.TIOCGWINSZ, '\0'*8))[:2]
			self.term_width = w

	def percentage(self):
		"""
			Returns the percentage of the progress.
		"""
		return self.currval*100.0 / self.maxval

	def _format_widgets(self):
		r = []
		hfill_inds = []
		num_hfill = 0
		currwidth = 0
		for (i, w) in enumerate(self.widgets):
			if isinstance(w, ProgressBarWidgetHFill):
				r.append(w)
				hfill_inds.append(i)
				num_hfill += 1
			elif isinstance(w, str):  # OR isinstance(w, (str, unicode))
				r.append(w)
				currwidth += len(w)
			else:
				weval = w.update()
				currwidth += len(weval)
				r.append(weval)
		for iw in hfill_inds:
			r[iw] = r[iw].update((self.term_width-currwidth)/num_hfill)
		return r

	def _format_line(self):
		return ''.join(self._format_widgets()).ljust(self.term_width)

	def _need_update(self):
		# return int(self.percentage()) != int(self.prev_percentage)
		return int(self.percentage() / self.update_step) != \
			int(self.prev_percentage / self.update_step)

	def update(self, value):
		"""
			Updates the progress bar to a new value.
		"""
		assert 0 <= value <= self.maxval
		self.currval = value
		if not self._need_update() or self.finished:
			return
		if not self.start_time:
			self.start_time = time.time()
		self.seconds_elapsed = time.time() - self.start_time
		self.prev_percentage = self.percentage()
		if value != self.maxval:
			self.fd.write(self._format_line() + '\r')
		else:
			self.finished = True
			self.fd.write(self._format_line() + '\n')

	def start(self):
		"""
			Start measuring time, and prints the bar at 0%.

			It returns self so you can use it like this:
			>>> pbar = ProgressBar().start()
			>>> for i in xrange(100):
			...    # do something
			...    pbar.update(i+1)
			...
			>>> pbar.finish()
		"""
		self.update(0)
		return self

	def finish(self):
		"""
			Used to tell the progress is finished.
		"""
		self.update(self.maxval)
		if self.signal_set:
			signal.signal(signal.SIGWINCH, signal.SIG_DFL)

if __name__ == '__main__':
	def example1():
		pbar = ProgressBar(
			widgets=[
				'Test: ',
				Bar(),
				' ',
				RotatingMarker(),
				Percentage(),
				' ',
				ETA(),
			],
			maxval=1.0,
			update_step=0.2,
		)
		pbar.start()
		for i in range(1000):
			# do something
			time.sleep(0.1)
			pbar.update(i/1000.0)
		pbar.finish()
		print('')
	example1()



# -*- coding: utf-8 -*-
# appledict/__init__.py
# Output to Apple Dictionary xml sources for Dictionary Development Kit.
#
# Copyright (C) 2016 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Copyright (C) 2016 Ratijas <ratijas.t@me.com>
# Copyright (C) 2012-2015 Xiaoqiang Wang <xiaoqiangwang AT gmail DOT com>
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

import sys
import os
from os.path import abspath, basename

import re
import pkgutil
import shutil

from pyglossary.plugins.formats_common import *
from ._dict import *

import xdxf

sys.setrecursionlimit(10000)

enable = True
format = "AppleDict"
description = "AppleDict Source (xml)"
extentions = [".xml"]
readOptions = []
writeOptions = [
	"cleanHTML",  # bool
	"css",  # str or None
	"xsl",  # str or None
	"defaultPrefs",  # dict or None, FIXME
	"prefsHTML",  # str or None
	"frontBackMatter",  # str or None
	"jing",  # str or None
	"indexes",  # str or None
]


def abspath_or_None(path):
	return os.path.abspath(os.path.expanduser(path)) if path else None


def write_header(glos, toFile, frontBackMatter):
	# write header
	toFile.write(
		'<?xml version="1.0" encoding="UTF-8"?>\n'
		'<d:dictionary xmlns="http://www.w3.org/1999/xhtml" '
		'xmlns:d="http://www.apple.com/DTDs/DictionaryService-1.0.rng">\n'
	)

	if frontBackMatter:
		with open(frontBackMatter, "r") as front_back_matter:
			toFile.write(front_back_matter.read())


def format_default_prefs(defaultPrefs):
	"""
	:type defaultPrefs: dict or None

	as by 14th of Jan 2016, it is highly recommended that prefs should contain
	{"version": "1"}, otherwise Dictionary.app does not keep user changes
	between restarts.
	"""
	if not defaultPrefs:
		return ""
	if not isinstance(defaultPrefs, dict):
		raise TypeError("defaultPrefs not a dictionary: %r" % defaultPrefs)
	if str(defaultPrefs.get("version", None)) != "1":
		log.error("default prefs does not contain {'version': '1'}.  prefs "
				  "will not be persistent between Dictionary.app restarts.")
	return "\n".join("\t\t<key>%s</key>\n\t\t<string>%s</string>" % i
					 for i in sorted(defaultPrefs.items())).strip()


def write_css(fname, css_file):
	with open(fname, "wb") as toFile:
		if css_file:
			with open(css_file, "rb") as fromFile:
				toFile.write(fromFile.read())
		else:
			toFile.write(pkgutil.get_data(
				__name__,
				"templates/Dictionary.css",
			))


def write(
	glos,
	dirPath,
	cleanHTML=True,
	css=None,
	xsl=None,
	defaultPrefs=None,
	prefsHTML=None,
	frontBackMatter=None,
	jing=None,
	indexes=None,
):
	"""
	write glossary to Apple dictionary .xml and supporting files.

	:type glos: pyglossary.glossary.Glossary
	:type dirPath: str, directory path, must not have extension

	:type cleanHTML: str
	:param cleanHTML: pass "yes" to use BeautifulSoup parser.

	:type css: str or None
	:param css: path to custom .css file

	:type xsl: str or None
	:param xsl: path to custom XSL transformations file.

	:type defaultPrefs: dict or None
	:param defaultPrefs: Default prefs in python dictionary literal format,
	i.e. {"key1": "value1", "key2": "value2", ...}.  All keys and values must
	be quoted strings; not allowed characters (e.g. single/double quotes,
	equal sign "=", semicolon) must be escaped as hex code according to
	python string literal rules.

	:type prefsHTML: str or None
	:param prefsHTML: path to XHTML file with user interface for dictionary's
	preferences.  refer to Apple's documentation for details.

	:type frontBackMatter: str or None
	:param frontBackMatter: path to XML file with top-level tag
	<d:entry id="front_back_matter" d:title="Your Front/Back Matter Title">
		your front/back matter entry content
	</d:entry>

	:type jing: str or None
	:param jing: pass "yes" to run Jing check on generated XML.

	:type indexes: str or None
	:param indexes: Dictionary.app is dummy and by default it don't know
	how to perform flexible search.  we can help it by manually providing
	additional indexes to dictionary entries.
	# for now no languages supported yet.
	"""
	if not isdir(dirPath):
		os.mkdir(dirPath)

	xdxf.xdxf_init()

	if cleanHTML:
		BeautifulSoup = get_beautiful_soup()
		if not BeautifulSoup:
			log.warning(
				"cleanHTML option passed but BeautifulSoup not found.  " +
				"to fix this run `sudo pip3 install lxml beautifulsoup4 html5lib`"
			)
	else:
		BeautifulSoup = None

	fileNameBase = basename(dirPath).replace(".", "_")
	filePathBase = join(dirPath, fileNameBase)
	# before chdir (outside indir block)
	css = abspath_or_None(css)
	xsl = abspath_or_None(xsl)
	prefsHTML = abspath_or_None(prefsHTML)
	frontBackMatter = abspath_or_None(frontBackMatter)

	generate_id = id_generator()
	generate_indexes = indexes_generator(indexes)

	glos.setDefaultDefiFormat("h")

	myResDir = join(dirPath, "OtherResources")
	if not isdir(myResDir):
		os.mkdir(myResDir)

	with open(filePathBase + ".xml", "w", encoding="utf8") as toFile:
		write_header(glos, toFile, frontBackMatter)
		for entryI, entry in enumerate(glos):
			if entry.isData():
				entry.save(myResDir)
				continue

			words = entry.getWords()
			word, alts = words[0], words[1:]
			defi = entry.getDefi()

			long_title = _normalize.title_long(
				_normalize.title(word, BeautifulSoup)
			)
			if not long_title:
				continue

			_id = next(generate_id)
			if BeautifulSoup:
				title_attr = BeautifulSoup.dammit.EntitySubstitution\
					.substitute_xml(long_title, True)
			else:
				title_attr = str(long_title)

			content_title = long_title
			if entry.getDefiFormat() == "x":
				defi = xdxf.xdxf_to_html(defi)
				content_title = None
			content = format_clean_content(content_title, defi, BeautifulSoup)

			toFile.write(
				'<d:entry id="%s" d:title=%s>\n' % (_id, title_attr) +
				generate_indexes(long_title, alts, content, BeautifulSoup) +
				content +
				"\n</d:entry>\n"
			)

		toFile.write("</d:dictionary>\n")

	if xsl:
		shutil.copy(xsl, myResDir)

	if prefsHTML:
		shutil.copy(prefsHTML, myResDir)

	write_css(filePathBase + ".css", css)

	with open(join(dirPath, "Makefile"), "w") as toFile:
		toFile.write(
			toStr(pkgutil.get_data(
				__name__,
				"templates/Makefile",
			)) % {"dict_name": fileNameBase}
		)

	copyright = glos.getInfo("copyright")
	if BeautifulSoup:
		# strip html tags
		copyright = str(BeautifulSoup.BeautifulSoup(
			copyright,
			"lxml"
		).text)

	# if DCSDictionaryXSL provided but DCSDictionaryDefaultPrefs <dict/> not
	# present in Info.plist, Dictionary.app will crash.
	with open(filePathBase + ".plist", "w", encoding="utf-8") as toFile:
		toFile.write(
			toStr(pkgutil.get_data(
				__name__,
				"templates/Info.plist",
			)) % {
				"CFBundleIdentifier":
					fileNameBase.replace(" ", ""),  # identifier must be unique
				"CFBundleDisplayName":
					glos.getInfo("name"),
				"CFBundleName":
					fileNameBase,
				"DCSDictionaryCopyright":
					copyright,
				"DCSDictionaryManufacturerName":
					glos.getInfo("author"),
				"DCSDictionaryXSL":
					basename(xsl) if xsl else "",
				"DCSDictionaryDefaultPrefs":
					format_default_prefs(defaultPrefs),
				"DCSDictionaryPrefsHTML":
					basename(prefsHTML) if prefsHTML else "",
				"DCSDictionaryFrontMatterReferenceID":
					"<key>DCSDictionaryFrontMatterReferenceID</key>\n"
					"\t<string>front_back_matter</string>" if frontBackMatter
					else "",
			}
		)

	if jing == "yes":
		from .jing import run as jing_run
		jing_run(filePathBase + ".xml")



# -*- coding: utf-8 -*-

from formats_common import *
from pyglossary.file_utils import fileCountLines

enable = True
format = "DictOrg"
description = "DICT.org file format (.index)"
extentions = [".index"]
readOptions = []
writeOptions = [
	"dictzip",  # bool
	"install",  # bool
]
sortOnWrite = DEFAULT_YES


b64_chars = b"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
b64_chars_ord = {c: i for i, c in enumerate(b64_chars)}


def intToIndexStr(n, retry=0):
	chars = []
	while True:
		chars.append(b64_chars[n & 0x3f])
		n >>= 6
		if n == 0:
			break
	return bytes(reversed(chars))


def indexStrToInt(st):
	n = 0
	for i, c in enumerate(reversed(list(st))):
		k = b64_chars_ord[c]
		assert 0 <= k < 64
		n |= (k << 6*i)
		# += is safe
		# |= is probably a little faster
		# |= is also safe because n has lesser that 6*i bits. why? ## FIXME
	return n


def installToDictd(filename, title=""):
	"""
	filename is without extention (neither .index or .dict or .dict.dz)
	"""
	import shutil
	targetDir = "/usr/share/dictd/"
	log.info("Installing %r to DICTD server" % filename)

	if os.path.isfile(filename + ".dict.dz"):
		dictPostfix = ".dict.dz"
	elif os.path.isfile(filename + ".dict"):
		dictPostfix = ".dict"
	else:
		log.error("No .dict file, could not install dictd file %r" % filename)
		return False

	if not filename.startswith(targetDir):
		shutil.copy(filename + ".index", targetDir)
		shutil.copy(filename + dictPostfix, targetDir)

	fname = split(filename)[1]
	if not title:
		title = fname
	open("/var/lib/dictd/db.list", "a").write("""
database %s
{
  data %s
  index %s
}
""" % (
	title,
	join(targetDir, fname + dictPostfix),
	join(targetDir, fname + ".index"),
))


class Reader(object):
	def __init__(self, glos):
		self._glos = glos
		self._filename = ""
		self._indexFp = None
		self._dictFp = None
		self._leadingLinesCount = 0
		self._len = None

	def open(self, filename):
		import gzip
		if filename.endswith(".index"):
			filename = filename[:-6]
		self._filename = filename
		self._indexFp = open(filename+".index", "rb")
		if os.path.isfile(filename+".dict.dz"):
			self._dictFp = gzip.open(filename+".dict.dz")
		else:
			self._dictFp = open(filename+".dict", "rb")

	def close(self):
		if self._indexFp is not None:
			try:
				self._indexFp.close()
			except:
				log.exception("error while closing index file")
			self._indexFp = None
		if self._dictFp is not None:
			try:
				self._dictFp.close()
			except:
				log.exception("error while closing dict file")
			self._dictFp = None
	def __len__(self):
		if self._len is None:
			log.debug("Try not to use len(reader) as it takes extra time")
			self._len = fileCountLines(
				self._filename + ".index"
			) - self._leadingLinesCount
		return self._len

	def __iter__(self):
		if not self._indexFp:
			log.error("reader is not open, can not iterate")
			raise StopIteration
		# read info from header of dict file # FIXME
		word = ""
		sumLen = 0
		wrongSortedN = 0
		wordCount = 0
		# __________________ IMPORTANT PART __________________ #
		for line in self._indexFp:
			line = line.strip()
			if not line:
				continue
			parts = line.split(b"\t")
			assert len(parts) == 3
			word = parts[0].replace(b"<BR>", b"\\n")\
						   .replace(b"<br>", b"\\n")
			sumLen2 = indexStrToInt(parts[1])
			if sumLen2 != sumLen:
				wrongSortedN += 1
			sumLen = sumLen2
			defiLen = indexStrToInt(parts[2])
			self._dictFp.seek(sumLen)
			defi = self._dictFp.read(defiLen)
			defi = defi.replace(b"<BR>", b"\n").replace(b"<br>", b"\n")
			sumLen += defiLen
			yield self._glos.newEntry(toStr(word), toStr(defi))
			wordCount += 1
		# ____________________________________________________ #

		if wrongSortedN > 0:
			log.warning("Warning: wrong sorting count: %d" % wrongSortedN)
		self._len = wordCount


def write(glos, filename, dictzip=True, install=True):  # FIXME
	from pyglossary.text_utils import runDictzip
	(filename_nox, ext) = splitext(filename)
	if ext.lower() == ".index":
		filename = filename_nox
	indexFd = open(filename+".index", "wb")
	dictFd = open(filename+".dict", "wb")
	dictMark = 0
	for entry in glos:
		if entry.isData():
			# does dictd support resources? and how? FIXME
			continue
		word = toBytes(entry.getWord())
		defi = toBytes(entry.getDefi())
		lm = len(defi)
		indexFd.write(
			word + b"\t" +
			intToIndexStr(dictMark) + b"\t" +
			intToIndexStr(lm) + b"\n"
		)  # FIXME
		dictFd.write(toBytes(defi))
		dictMark += lm
	indexFd.close()
	dictFd.close()
	# for key, value in glos.iterInfo():
	#	if not value:
	#		continue
	#	pass  # FIXME
	if dictzip:
		runDictzip(filename)
	if install:
		installToDictd(filename, glos.getInfo("name").replace(" ", "_"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# readmdict.py
# Octopus MDict Dictionary File (.mdx) and Resource File (.mdd) Analyser
#
# Copyright (C) 2012, 2013, 2015 Xiaoqiang Wang <xiaoqiangwang AT gmail DOT com>
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

import logging
log = logging.getLogger('root')

from struct import pack, unpack
from io import BytesIO
import re
import sys

from .ripemd128 import ripemd128
from .pureSalsa20 import Salsa20

# zlib compression is used for engine version >=2.0
import zlib
# LZO compression is used for engine version < 2.0
try:
	import lzo
except ImportError:
	lzo = None
	print("LZO compression support is not available")

# 2x3 compatible
if sys.hexversion >= 0x03000000:
	str, unicode = bytes, str


def _unescape_entities(text):
	"""
	unescape offending tags < > " &
	"""
	text = text.replace(b'&lt;', b'<')
	text = text.replace(b'&gt;', b'>')
	text = text.replace(b'&quot;', b'"')
	text = text.replace(b'&amp;', b'&')
	return text


def _fast_decrypt(data, key):
	b = bytearray(data)
	key = bytearray(key)
	previous = 0x36
	for i in range(len(b)):
		t = (b[i] >> 4 | b[i] << 4) & 0xff
		t = t ^ previous ^ (i & 0xff) ^ key[i % len(key)]
		previous = b[i]
		b[i] = t
	return bytes(b)


def _mdx_decrypt(comp_block):
	key = ripemd128(comp_block[4:8] + pack(b'<L', 0x3695))
	return comp_block[0:8] + _fast_decrypt(comp_block[8:], key)


def _salsa_decrypt(ciphertext, encrypt_key):
	s20 = Salsa20(key=encrypt_key, IV=b"\x00"*8, rounds=8)
	return s20.encryptBytes(ciphertext)


def _decrypt_regcode_by_deviceid(reg_code, deviceid):
	deviceid_digest = ripemd128(deviceid)
	s20 = Salsa20(key=deviceid_digest, IV=b"\x00"*8, rounds=8)
	encrypt_key = s20.encryptBytes(reg_code)
	return encrypt_key


def _decrypt_regcode_by_email(reg_code, email):
	email_digest = ripemd128(email.decode().encode('utf-16-le'))
	s20 = Salsa20(key=email_digest, IV=b"\x00"*8, rounds=8)
	encrypt_key = s20.encryptBytes(reg_code)
	return encrypt_key


class MDict(object):
	"""
	Base class which reads in header and key block.
	It has no public methods and serves only as code sharing base class.
	"""
	def __init__(self, fname, encoding='', passcode=None):
		self._fname = fname
		self._encoding = encoding.upper()
		self._passcode = passcode

		self.header = self._read_header()
		try:
			self._key_list = self._read_keys()
		except:
			print("Try Brutal Force on Encrypted Key Blocks")
			self._key_list = self._read_keys_brutal()

	def __len__(self):
		return self._num_entries

	def __iter__(self):
		return self.keys()

	def keys(self):
		"""
		Return an iterator over dictionary keys.
		"""
		return (key_value for key_id, key_value in self._key_list)

	def _read_number(self, f):
		return unpack(self._number_format, f.read(self._number_width))[0]

	def _parse_header(self, header):
		"""
		extract attributes from <Dict attr="value" ... >
		"""
		taglist = re.findall(b'(\w+)="(.*?)"', header, re.DOTALL)
		tagdict = {}
		for key, value in taglist:
			tagdict[key] = _unescape_entities(value)
		return tagdict

	def _decode_key_block_info(self, key_block_info_compressed):
		if self._version >= 2:
			# zlib compression
			assert(key_block_info_compressed[:4] == b'\x02\x00\x00\x00')
			# decrypt if needed
			if self._encrypt & 0x02:
				key_block_info_compressed = _mdx_decrypt(key_block_info_compressed)
			# decompress
			key_block_info = zlib.decompress(key_block_info_compressed[8:])
			# adler checksum
			adler32 = unpack('>I', key_block_info_compressed[4:8])[0]
			assert(adler32 == zlib.adler32(key_block_info) & 0xffffffff)
		else:
			# no compression
			key_block_info = key_block_info_compressed
		# decode
		key_block_info_list = []
		num_entries = 0
		i = 0
		if self._version >= 2:
			byte_format = '>H'
			byte_width = 2
			text_term = 1
		else:
			byte_format = '>B'
			byte_width = 1
			text_term = 0

		while i < len(key_block_info):
			# number of entries in current key block
			num_entries += unpack(self._number_format, key_block_info[i:i+self._number_width])[0]
			i += self._number_width
			# text head size
			text_head_size = unpack(byte_format, key_block_info[i:i+byte_width])[0]
			i += byte_width
			# text head
			if self._encoding != 'UTF-16':
				i += text_head_size + text_term
			else:
				i += (text_head_size + text_term) * 2
			# text tail size
			text_tail_size = unpack(byte_format, key_block_info[i:i+byte_width])[0]
			i += byte_width
			# text tail
			if self._encoding != 'UTF-16':
				i += text_tail_size + text_term
			else:
				i += (text_tail_size + text_term) * 2
			# key block compressed size
			key_block_compressed_size = unpack(self._number_format, key_block_info[i:i+self._number_width])[0]
			i += self._number_width
			# key block decompressed size
			key_block_decompressed_size = unpack(self._number_format, key_block_info[i:i+self._number_width])[0]
			i += self._number_width
			key_block_info_list += [(key_block_compressed_size, key_block_decompressed_size)]

		#assert(num_entries == self._num_entries)

		return key_block_info_list

	def _decode_key_block(self, key_block_compressed, key_block_info_list):
		key_list = []
		i = 0
		for compressed_size, decompressed_size in key_block_info_list:
			start = i
			end = i + compressed_size
			# 4 bytes : compression type
			key_block_type = key_block_compressed[start:start+4]
			# 4 bytes : adler checksum of decompressed key block
			adler32 = unpack('>I', key_block_compressed[start+4:start+8])[0]
			if key_block_type == b'\x00\x00\x00\x00':
				key_block = key_block_compressed[start+8:end]
			elif key_block_type == b'\x01\x00\x00\x00':
				if lzo is None:
					print("LZO compression is not supported")
					break
				# decompress key block
				header = b'\xf0' + pack('>I', decompressed_size)
				key_block = lzo.decompress(header + key_block_compressed[start+8:end])
			elif key_block_type == b'\x02\x00\x00\x00':
				# decompress key block
				key_block = zlib.decompress(key_block_compressed[start+8:end])
			# extract one single key block into a key list
			key_list += self._split_key_block(key_block)
			# notice that adler32 returns signed value
			assert(adler32 == zlib.adler32(key_block) & 0xffffffff)

			i += compressed_size
		return key_list

	def _split_key_block(self, key_block):
		key_list = []
		key_start_index = 0
		while key_start_index < len(key_block):
			# the corresponding record's offset in record block
			key_id = unpack(self._number_format, key_block[key_start_index:key_start_index+self._number_width])[0]
			# key text ends with '\x00'
			if self._encoding == 'UTF-16':
				delimiter = b'\x00\x00'
				width = 2
			else:
				delimiter = b'\x00'
				width = 1
			i = key_start_index + self._number_width
			while i < len(key_block):
				if key_block[i:i+width] == delimiter:
					key_end_index = i
					break
				i += width
			key_text = key_block[key_start_index+self._number_width:key_end_index]\
				.decode(self._encoding, errors='ignore').encode('utf-8').strip()
			key_start_index = key_end_index + width
			key_list += [(key_id, key_text)]
		return key_list

	def _read_header(self):
		f = open(self._fname, 'rb')
		# number of bytes of header text
		header_bytes_size = unpack('>I', f.read(4))[0]
		header_bytes = f.read(header_bytes_size)
		# 4 bytes: adler32 checksum of header, in little endian
		adler32 = unpack('<I', f.read(4))[0]
		assert(adler32 == zlib.adler32(header_bytes) & 0xffffffff)
		# mark down key block offset
		self._key_block_offset = f.tell()
		f.close()

		# header text in utf-16 encoding ending with '\x00\x00'
		header_text = header_bytes[:-2].decode('utf-16').encode('utf-8')
		header_tag = self._parse_header(header_text)
		if not self._encoding:
			encoding = header_tag[b'Encoding']
			if sys.hexversion >= 0x03000000:
				encoding = encoding.decode('utf-8')
			# GB18030 > GBK > GB2312
			if encoding in ['GBK', 'GB2312']:
				encoding = 'GB18030'
			self._encoding = encoding
		# encryption flag
		#   0x00 - no encryption
		#   0x01 - encrypt record block
		#   0x02 - encrypt key info block
		if b'Encrypted' not in header_tag or header_tag[b'Encrypted'] == b'No':
			self._encrypt = 0
		elif header_tag[b'Encrypted'] == b'Yes':
			self._encrypt = 1
		else:
			self._encrypt = int(header_tag[b'Encrypted'])

		# stylesheet attribute if present takes form of:
		#   style_number # 1-255
		#   style_begin  # or ''
		#   style_end    # or ''
		# store stylesheet in dict in the form of
		# {'number' : ('style_begin', 'style_end')}
		self._stylesheet = {}
		if header_tag.get('StyleSheet'):
			lines = header_tag['StyleSheet'].splitlines()
			for i in range(0, len(lines), 3):
				self._stylesheet[lines[i]] = (lines[i+1], lines[i+2])

		# before version 2.0, number is 4 bytes integer
		# version 2.0 and above uses 8 bytes
		self._version = float(header_tag[b'GeneratedByEngineVersion'])
		if self._version < 2.0:
			self._number_width = 4
			self._number_format = '>I'
		else:
			self._number_width = 8
			self._number_format = '>Q'

		return header_tag

	def _read_keys(self):
		f = open(self._fname, 'rb')
		f.seek(self._key_block_offset)

		# the following numbers could be encrypted
		if self._version >= 2.0:
			num_bytes = 8 * 5
		else:
			num_bytes = 4 * 4
		block = f.read(num_bytes)

		if self._encrypt & 1:
			if self._passcode is None:
				raise RuntimeError('user identification is needed to read encrypted file')
			regcode, userid = self._passcode
			if isinstance(userid, unicode):
				userid = userid.encode('utf8')
			if self.header[b'RegisterBy'] == b'EMail':
				encrypted_key = _decrypt_regcode_by_email(regcode, userid)
			else:
				encrypted_key = _decrypt_regcode_by_deviceid(regcode, userid)
			block = _salsa_decrypt(block, encrypted_key)

		# decode this block
		sf = BytesIO(block)
		# number of key blocks
		num_key_blocks = self._read_number(sf)
		# number of entries
		self._num_entries = self._read_number(sf)
		# number of bytes of key block info after decompression
		if self._version >= 2.0:
			key_block_info_decomp_size = self._read_number(sf)
		# number of bytes of key block info
		key_block_info_size = self._read_number(sf)
		# number of bytes of key block
		key_block_size = self._read_number(sf)

		# 4 bytes: adler checksum of previous 5 numbers
		if self._version >= 2.0:
			adler32 = unpack('>I', f.read(4))[0]
			assert adler32 == (zlib.adler32(block) & 0xffffffff)

		# read key block info, which indicates key block's compressed and decompressed size
		key_block_info = f.read(key_block_info_size)
		key_block_info_list = self._decode_key_block_info(key_block_info)
		assert(num_key_blocks == len(key_block_info_list))

		# read key block
		key_block_compressed = f.read(key_block_size)
		# extract key block
		key_list = self._decode_key_block(key_block_compressed, key_block_info_list)

		self._record_block_offset = f.tell()
		f.close()

		return key_list

	def _read_keys_brutal(self):
		f = open(self._fname, 'rb')
		f.seek(self._key_block_offset)

		# the following numbers could be encrypted, disregard them!
		if self._version >= 2.0:
			num_bytes = 8 * 5 + 4
			key_block_type = b'\x02\x00\x00\x00'
		else:
			num_bytes = 4 * 4
			key_block_type = b'\x01\x00\x00\x00'
		block = f.read(num_bytes)

		# key block info
		# 4 bytes '\x02\x00\x00\x00'
		# 4 bytes adler32 checksum
		# unknown number of bytes follows until '\x02\x00\x00\x00' which marks the beginning of key block
		key_block_info = f.read(8)
		if self._version >= 2.0:
			assert key_block_info[:4] == b'\x02\x00\x00\x00'
		while True:
			fpos = f.tell()
			t = f.read(1024)
			index = t.find(key_block_type)
			if index != -1:
				key_block_info += t[:index]
				f.seek(fpos + index)
				break
			else:
				key_block_info += t

		key_block_info_list = self._decode_key_block_info(key_block_info)
		key_block_size = sum(list(zip(*key_block_info_list))[0])

		# read key block
		key_block_compressed = f.read(key_block_size)
		# extract key block
		key_list = self._decode_key_block(key_block_compressed, key_block_info_list)

		self._record_block_offset = f.tell()
		f.close()

		self._num_entries = len(key_list)
		return key_list


class MDD(MDict):
	"""
	MDict resource file format (*.MDD) reader.
	>>> mdd = MDD('example.mdd')
	>>> len(mdd)
	208
	>>> for filename,content in mdd.items():
	... print filename, content[:10]
	"""
	def __init__(self, fname, passcode=None):
		MDict.__init__(self, fname, encoding='UTF-16', passcode=passcode)

	def items(self):
		"""Return a generator which in turn produce tuples in the form of (filename, content)
		"""
		return self._decode_record_block()

	def _decode_record_block(self):
		f = open(self._fname, 'rb')
		f.seek(self._record_block_offset)

		num_record_blocks = self._read_number(f)
		num_entries = self._read_number(f)
		assert(num_entries == self._num_entries)
		record_block_info_size = self._read_number(f)
		record_block_size = self._read_number(f)

		# record block info section
		record_block_info_list = []
		size_counter = 0
		for i in range(num_record_blocks):
			compressed_size = self._read_number(f)
			decompressed_size = self._read_number(f)
			record_block_info_list += [(compressed_size, decompressed_size)]
			size_counter += self._number_width * 2
		assert(size_counter == record_block_info_size)

		# actual record block
		offset = 0
		i = 0
		size_counter = 0
		for compressed_size, decompressed_size in record_block_info_list:
			record_block_compressed = f.read(compressed_size)
			# 4 bytes: compression type
			record_block_type = record_block_compressed[:4]
			# 4 bytes: adler32 checksum of decompressed record block
			adler32 = unpack('>I', record_block_compressed[4:8])[0]
			if record_block_type == b'\x00\x00\x00\x00':
				record_block = record_block_compressed[8:]
			elif record_block_type == b'\x01\x00\x00\x00':
				if lzo is None:
					print("LZO compression is not supported")
					break
				# decompress
				header = '\xf0' + pack('>I', decompressed_size)
				record_block = lzo.decompress(header + record_block_compressed[8:])
			elif record_block_type == b'\x02\x00\x00\x00':
				# decompress
				record_block = zlib.decompress(record_block_compressed[8:])

			# notice that adler32 return signed value
			assert(adler32 == zlib.adler32(record_block) & 0xffffffff)

			assert(len(record_block) == decompressed_size)
			# split record block according to the offset info from key block
			while i < len(self._key_list):
				record_start, key_text = self._key_list[i]
				# reach the end of current record block
				if record_start - offset >= len(record_block):
					break
				# record end index
				if i < len(self._key_list)-1:
					record_end = self._key_list[i+1][0]
				else:
					record_end = len(record_block) + offset
				i += 1
				data = record_block[record_start-offset:record_end-offset]
				yield key_text, data
			offset += len(record_block)
			size_counter += compressed_size
		assert(size_counter == record_block_size)

		f.close()


class MDX(MDict):
	"""
	MDict dictionary file format (*.MDD) reader.
	>>> mdx = MDX('example.mdx')
	>>> len(mdx)
	42481
	>>> for key,value in mdx.items():
	... print key, value[:10]
	"""
	def __init__(self, fname, encoding='', substyle=False, passcode=None):
		MDict.__init__(self, fname, encoding, passcode)
		self._substyle = substyle

	def items(self):
		"""Return a generator which in turn produce tuples in the form of (key, value)
		"""
		return self._decode_record_block()

	def _substitute_stylesheet(self, txt):
		# substitute stylesheet definition
		txt_list = re.split('`\d+`', txt)
		txt_tag = re.findall('`\d+`', txt)
		txt_styled = txt_list[0]
		for j, p in enumerate(txt_list[1:]):
			key = txt_tag[j][1:-1]
			try:
				style = self._stylesheet[key]
			except KeyError:
				log.error('invalid stylesheet key "%s"'%key)
				continue
			if p and p[-1] == '\n':
				txt_styled = txt_styled + style[0] + p.rstrip() + style[1] + '\r\n'
			else:
				txt_styled = txt_styled + style[0] + p + style[1]
		return txt_styled

	def _decode_record_block(self):
		f = open(self._fname, 'rb')
		f.seek(self._record_block_offset)

		num_record_blocks = self._read_number(f)
		num_entries = self._read_number(f)
		assert(num_entries == self._num_entries)
		record_block_info_size = self._read_number(f)
		record_block_size = self._read_number(f)

		# record block info section
		record_block_info_list = []
		size_counter = 0
		for i in range(num_record_blocks):
			compressed_size = self._read_number(f)
			decompressed_size = self._read_number(f)
			record_block_info_list += [(compressed_size, decompressed_size)]
			size_counter += self._number_width * 2
		assert(size_counter == record_block_info_size)

		# actual record block data
		offset = 0
		i = 0
		size_counter = 0
		for compressed_size, decompressed_size in record_block_info_list:
			record_block_compressed = f.read(compressed_size)
			# 4 bytes indicates block compression type
			record_block_type = record_block_compressed[:4]
			# 4 bytes adler checksum of uncompressed content
			adler32 = unpack('>I', record_block_compressed[4:8])[0]
			# no compression
			if record_block_type == b'\x00\x00\x00\x00':
				record_block = record_block_compressed[8:]
			# lzo compression
			elif record_block_type == b'\x01\x00\x00\x00':
				if lzo is None:
					print("LZO compression is not supported")
					break
				# decompress
				header = b'\xf0' + pack('>I', decompressed_size)
				record_block = lzo.decompress(header + record_block_compressed[8:])
			# zlib compression
			elif record_block_type == b'\x02\x00\x00\x00':
				# decompress
				record_block = zlib.decompress(record_block_compressed[8:])

			# notice that adler32 return signed value
			assert(adler32 == zlib.adler32(record_block) & 0xffffffff)

			assert(len(record_block) == decompressed_size)
			# split record block according to the offset info from key block
			while i < len(self._key_list):
				record_start, key_text = self._key_list[i]
				# reach the end of current record block
				if record_start - offset >= len(record_block):
					break
				# record end index
				if i < len(self._key_list)-1:
					record_end = self._key_list[i+1][0]
				else:
					record_end = len(record_block) + offset
				i += 1
				record = record_block[record_start-offset:record_end-offset]
				# convert to utf-8
				record = record.decode(self._encoding, errors='ignore').strip(unicode('\x00')).encode('utf-8')
				# substitute styles
				if self._substyle and self._stylesheet:
					record = self._substitute_stylesheet(record)

				yield key_text, record
			offset += len(record_block)
			size_counter += compressed_size
		assert(size_counter == record_block_size)

		f.close()


if __name__ == '__main__':
	import sys
	import os
	import os.path
	import argparse
	import codecs

	def passcode(s):
		try:
			regcode, userid = s.split(',')
		except:
			raise argparse.ArgumentTypeError("Passcode must be regcode,userid")
		try:
			regcode = codecs.decode(regcode, 'hex')
		except:
			raise argparse.ArgumentTypeError("regcode must be a 32 bytes hexadecimal string")
		return regcode, userid

	parser = argparse.ArgumentParser()
	parser.add_argument('-x', '--extract', action="store_true",
						help='extract mdx to source format and extract files from mdd')
	parser.add_argument('-s', '--substyle', action="store_true",
						help='substitute style definition if present')
	parser.add_argument('-d', '--datafolder', default="data",
						help='folder to extract data files from mdd')
	parser.add_argument('-e', '--encoding', default="",
						help='folder to extract data files from mdd')
	parser.add_argument('-p', '--passcode', default=None, type=passcode,
						help='register_code,email_or_deviceid')
	parser.add_argument("filename", nargs='?', help="mdx file name")
	args = parser.parse_args()

	# use GUI to select file, default to extract
	if not args.filename:
		if sys.hexversion >= 0x03000000:
			import tkinter as tk
			import tkinter.filedialog as filedialog
		else:
			import Tkinter as tk
			import tkFileDialog as filedialog

		root = tk.Tk()
		root.withdraw()
		args.filename = filedialog.askopenfilename(parent=root)
		args.extract = True

	if not os.path.exists(args.filename):
		print("Please specify a valid MDX/MDD file")

	base, ext = os.path.splitext(args.filename)

	# read mdx file
	if ext.lower() == os.path.extsep + 'mdx':
		mdx = MDX(args.filename, args.encoding, args.substyle, args.passcode)
		if type(args.filename) is unicode:
			bfname = args.filename.encode('utf-8')
		else:
			bfname = args.filename
		print('======== %s ========' % bfname)
		print('  Number of Entries : %d' % len(mdx))
		for key, value in mdx.header.items():
			print('  %s : %s' % (key, value))
	else:
		mdx = None

	# find companion mdd file
	mdd_filename = ''.join([base, os.path.extsep, 'mdd'])
	if os.path.exists(mdd_filename):
		mdd = MDD(mdd_filename, args.passcode)
		if type(mdd_filename) is unicode:
			bfname = mdd_filename.encode('utf-8')
		else:
			bfname = mdd_filename
		print('======== %s ========' % bfname)
		print('  Number of Entries : %d' % len(mdd))
		for key, value in mdd.header.items():
			print('  %s : %s' % (key, value))
	else:
		mdd = None

	if args.extract:
		# write out glos
		if mdx:
			output_fname = ''.join([base, os.path.extsep, 'txt'])
			tf = open(output_fname, 'wb')
			for key, value in mdx.items():
				tf.write(key)
				tf.write(b'\r\n')
				tf.write(value)
				if not value.endswith(b'\n'):
					tf.write(b'\r\n')
				tf.write(b'</>\r\n')
			tf.close()
			# write out style
			if mdx.header.get('StyleSheet'):
				style_fname = ''.join([base, '_style', os.path.extsep, 'txt'])
				sf = open(style_fname, 'wb')
				sf.write(b'\r\n'.join(mdx.header['StyleSheet'].splitlines()))
				sf.close()
		# write out optional data files
		if mdd:
			datafolder = os.path.join(os.path.dirname(args.filename), args.datafolder)
			if not os.path.exists(datafolder):
				os.makedirs(datafolder)
			for key, value in mdd.items():
				fname = key.decode('utf-8').replace('\\', os.path.sep)
				dfname = datafolder + fname
				if not os.path.exists(os.path.dirname(dfname)):
					os.makedirs(os.path.dirname(dfname))
				df = open(dfname, 'wb')
				df.write(value)
				df.close()



from . import *
from os.path import isabs, join
from pyglossary.core import appResDir


def set_tooltip(widget, text):
	try:
		widget.set_tooltip_text(text)  # PyGTK 2.12 or above
	except AttributeError:
		try:
			widget.set_tooltip(gtk.Tooltips(), text)
		except:
			myRaise(__file__)


def imageFromFile(path):  # the file must exist
	if not isabs(path):
		path = join(appResDir, path)
	im = gtk.Image()
	try:
		im.set_from_file(path)
	except:
		myRaise()
	return im



import os
import shutil


class indir(object):
	"""
	mkdir + chdir shortcut to use with `with` statement.

		>>> print(os.getcwd())  # -> "~/projects"
		>>> with indir('my_directory', create=True):
		>>>     print(os.getcwd())  # -> "~/projects/my_directory"
		>>>     # do some work inside new 'my_directory'...
		>>> print(os.getcwd())  # -> "~/projects"
		>>> # automatically return to previous directory.
	"""
	def __init__(self, directory, create=False, clear=False):
		self.oldpwd = None
		self.dir = directory
		self.create = create
		self.clear = clear

	def __enter__(self):
		self.oldpwd = os.getcwd()
		if os.path.exists(self.dir):
			if self.clear:
				shutil.rmtree(self.dir)
				os.makedirs(self.dir)
		elif self.create:
			os.makedirs(self.dir)
		os.chdir(self.dir)

	def __exit__(self, exc_type, exc_val, exc_tb):
		os.chdir(self.oldpwd)
		self.oldpwd = None


def my_url_show(link):
	import subprocess
	for path in (
		'/usr/bin/gnome-www-browser',
		'/usr/bin/firefox',
		'/usr/bin/iceweasel',
		'/usr/bin/konqueror',
	):
		if os.path.isfile(path):
			subprocess.call([path, link])
			break
"""
try:
	from gnome import url_show
except:
	try:
		from gnomevfs import url_show
	except:
		url_show = my_url_show
"""


def click_website(widget, link):
	my_url_show(link)



# -*- coding: utf-8 -*-
# flawless_dsl/__init__.py
#
#
# Copyright (C) 2016 Ratijas <ratijas.t@me.com>
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
"""
only clean flawless DSL markup on output!
"""

from . import layer
from . import tag
from .main import (
	FlawlessDSLParser,
	parse,
)



# -*- coding: utf-8 -*-
#
# Copyright © 2008-2016 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Copyright © 2011-2012 kubtek <kubtek@gmail.com>
# This file is part of PyGlossary project, http://github.com/ilius/pyglossary
# Thanks to Raul Fernandes <rgfbr@yahoo.com.br> and Karl Grill
#	   for reverse engineering
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

from formats_common import *
from .bgl_reader import BglReader as Reader
from .bgl_reader import readOptions

enable = True
format = "BabylonBgl"
description = "Babylon (bgl)"
extentions = [".bgl"]

writeOptions = []
supportsAlternates = True
# progressbar = DEFAULT_YES

# FIXME: document type of read/write options
# (that would be specified in command line)



# from xml.sax.saxutils import escape as xml_escape
# from xml.sax.saxutils import unescape as xml_unescape


def xml_escape(data, entities=None):
	"""Escape &, <, and > in a string of data.

	You can escape other strings of data by passing a dictionary as
	the optional entities parameter.  The keys and values must all be
	strings; each key will be replaced with its corresponding value.
	"""
	if entities is None:
		entities = {}

	# must do ampersand first
	data = data.replace("&", "&amp;")
	data = data.replace(">", "&gt;")
	data = data.replace("<", "&lt;")
	if entities:
		data = __dict_replace(data, entities)
	return data


def xml_unescape(data, entities=None):
	"""Unescape &amp;, &lt;, and &gt; in a string of data.

	You can unescape other strings of data by passing a dictionary as
	the optional entities parameter.  The keys and values must all be
	strings; each key will be replaced with its corresponding value.
	"""
	if entities is None:
		entities = {}

	data = data.replace("&lt;", "<")
	data = data.replace("&gt;", ">")
	if entities:
		data = __dict_replace(data, entities)
	# must do ampersand last
	return data.replace("&amp;", "&")



# -*- coding: utf-8 -*-
# text_utils.py
#
# Copyright © 2008-2010 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

import string
import sys
import os
import re
import logging

from . import core

log = logging.getLogger("root")

startRed = "\x1b[31m"
endFormat = "\x1b[0;0;0m"  # len=8


def toBytes(s):
	return bytes(s, "utf8") if isinstance(s, str) else bytes(s)


def toStr(s):
	return str(s, "utf8") if isinstance(s, bytes) else str(s)


def fixUtf8(st):
	return toBytes(st).replace(b"\x00", b"").decode("utf-8", "replace")

pattern_n_us = re.compile(r"((?<!\\)(?:\\\\)*)\\n")
pattern_t_us = re.compile(r"((?<!\\)(?:\\\\)*)\\t")
pattern_bar_us = re.compile(r"((?<!\\)(?:\\\\)*)\\\|")
pattern_bar_sp = re.compile(r"(?:(?<!\\)(?:\\\\)*)\|")


def escapeNTB(st, bar=True):
	"""
		scapes Newline, Tab, Baskslash, and vertical Bar (if bar=True)
	"""
	st = st.replace(r"\\", r"\\\\")
	st = st.replace("\t", r"\t")
	st = st.replace("\n", r"\n")
	if bar:
		st = st.replace("|", r"\|")
	return st


def unescapeNTB(st, bar=False):
	"""
		unscapes Newline, Tab, Baskslash, and vertical Bar (if bar=True)
	"""
	st = re.sub(pattern_n_us, "\\1\n", st)
	st = re.sub(pattern_t_us, "\\1\t", st)
	if bar:
		st = re.sub(pattern_bar_us, r"\1\|", st)
#	st = re.sub(r"\\\\", r"\\", st)
	st = st.replace("\\\\", "\\")  # probably faster than re.sub
	return st


def splitByBarUnescapeNTB(st):
	"""
		splits by "|" (and not "\\|") then unescapes Newline (\\n),
			Tab (\\t), Baskslash (\\) and Bar (\\|) in each part
		returns a list
	"""
	return [
		unescapeNTB(part, bar=True)
		for part in re.split(pattern_bar_sp, st)
	]


# return a message string describing the current exception
def excMessage():
	i = sys.exc_info()
	return "{0}: {1}".format(i[0].__name__, i[1])


def formatHMS(h, m, s):
	if h == 0:
		if m == 0:
			return "%.2d" % s
		else:
			return "%.2d:%.2d" % (m, s)
	else:
		return "%.2d:%.2d:%.2d" % (h, m, s)


def timeHMS(seconds):
	import time
	(h, m, s) = time.gmtime(int(seconds))[3:6]
	return formatHMS(h, m, s)


def relTimeHMS(seconds):
	(days, s) = divmod(int(seconds), 24*3600)
	(m, s) = divmod(s, 60)
	(h, m) = divmod(m, 60)
	return formatHMS(h, m, s)

# ___________________________________________ #


def intToBinStr(n, stLen=0):
	bs = []
	while n > 0:
		bs.insert(0, n & 0xff)
		n >>= 8
	return bytes(bs).rjust(stLen, b"\x00")


def binStrToInt(bs):
	bs = toBytes(bs)
	n = 0
	for c in bs:
		n = (n << 8) + c
	return n


# ___________________________________________ #

def urlToPath(url):
	if not url.startswith("file://"):
		return url
	path = url[7:]
	if path[-2:] == "\r\n":
		path = path[:-2]
	elif path[-1] == "\r":
		path = path[:-1]
	# here convert html unicode symbols to utf8 string:
	if "%" not in path:
		return path
	path2 = ""
	n = len(path)
	i = 0
	while i < n:
		if path[i] == "%" and i < n-2:
			path2 += chr(eval("0x%s" % path[i+1:i+3]))
			i += 3
		else:
			path2 += path[i]
			i += 1
	return path2


def replacePostSpaceChar(st, ch):
	return st.replace(" "+ch, ch).replace(ch, ch+" ").replace(ch+"  ", ch+" ")


def runDictzip(filename):
	import subprocess
	dictzipCmd = "/usr/bin/dictzip"  # Save in pref FIXME
	if not os.path.isfile(dictzipCmd):
		return False
	if filename[-4:] == ".ifo":
		filename = filename[:-4]
	(out, err) = subprocess.Popen(
		[dictzipCmd, filename+".dict"],
		stdout=subprocess.PIPE
	).communicate()
#	out = p3[1].read()
#	err = p3[2].read()
#	log.debug("dictzip command: \"%s\""%dictzipCmd)
#	if err:
#		log.error("dictzip error: %s"%err.replace("\n", " "))
#	if out:
#		log.error("dictzip error: %s"%out.replace("\n", " "))


def isControlChar(y):
	# y: char code
	if y < 32 and chr(y) not in "\t\n\r\v":
		return True
	# according to ISO-8859-1
	if 128 <= y <= 159:
		return True
	return False


def isASCII(data, exclude=None):
	if exclude is None:
		exclude = []
	for c in data:
		co = ord(c)
		if co >= 128 and co not in exclude:
			return False
	return True


def formatByteStr(text):
	out = ""
	for c in text:
		out += "{0:0>2x}".format(ord(c)) + " "
	return out



def chBaseIntToList(number, base):
	result = []
	if number < 0:
		raise ValueError('number must be posotive integer')
	while True:
		number, rdigit = divmod(number, base)
		result = [rdigit] + result
		if number == 0:
			return result



import re
from formats_common import log
from . import conv

enable = True
format = "CC-CEDICT"
description = "CC-CEDICT"
# this typo is part of the API used by PyGlossary; don't change it!
extentions = (".u8",)

entry_count_reg = re.compile(r"#! entries=(\d+)")
class Reader:
	def __init__(self, glos):
		self._glos = glos
		self.file = None
		self.total_entries = self.entries_left = None

	def open(self, filename, encoding="utf-8"):
		if self.file is not None:
			self.file.close()
		self.file = open(filename, "r", encoding=encoding)
		for line in self.file:
			match = entry_count_reg.match(line)
			if match is not None:
				count = match.groups()[0]
				self.total_entries = self.entries_left = int(count)
				break
		else:
			self.close()
			raise RuntimeError("CC-CEDICT: could not find entry count")

	def close(self):
		if self.file is not None:
			self.file.close()
		self.file = None
		self.total_entries = self.entries_left = None

	def __len__(self):
		if self.total_entries is None:
			raise RuntimeError("CC-CEDICT: len(reader) called "\
					"while reader is not open")
		return self.total_entries

	def __iter__(self):
		if self.file is None:
			raise RuntimeError("CC-CEDICT: tried to iterate over entries "\
					"while reader is not open")
		for line in self.file:
			if not line.startswith("#"):
				if self.entries_left == 0:
					log.warning("more entries than the header claimed?!")
				self.entries_left -= 1
				parts = conv.parse_line(line)
				if parts is None:
					log.warning("bad line: %s", line)
					continue
				names, article = conv.make_entry(*parts)
				entry = self._glos.newEntry(names, article, defiFormat="h")
				yield entry



# -*- coding: utf-8 -*-
#
# Copyright © 2008-2016 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Copyright © 2011-2012 kubtek <kubtek@gmail.com>
# This file is part of PyGlossary project, http://github.com/ilius/pyglossary
# Thanks to Raul Fernandes <rgfbr@yahoo.com.br> and Karl Grill
#	   for reverse engineering
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

from .bgl_language import languageByCode
from .bgl_charset import charsetByCode

from pyglossary.plugins.formats_common import log
import pyglossary.gregorian as gregorian
from pyglossary.text_utils import (
	binStrToInt,
)


def decodeBglBinTime(b_value):
	jd1970 = gregorian.to_jd(1970, 1, 1)
	djd, hm = divmod(binStrToInt(b_value), 24*60)
	year, month, day = gregorian.jd_to(djd + jd1970)
	hour, minute = divmod(hm, 60)
	return "%.2d/%.2d/%.2d, %.2d:%.2d" % (year, month, day, hour, minute)


def languageInfoDecode(b_value):
	"""
		returns BabylonLanguage instance
	"""
	intValue = binStrToInt(b_value)
	try:
		return languageByCode[intValue]
	except IndexError:
		log.warning("read_type_3: unknown language code = %s" % intValue)
		return


def charsetInfoDecode(b_value):
	value = b_value[0]
	try:
		return charsetByCode[value]
	except KeyError:
		log.warning("read_type_3: unknown charset %s" % value)


def aboutInfoDecode(b_value):
	if not b_value:
		return
	aboutExt, _, aboutContents = b_value.partition(b"\x00")
	if not aboutExt:
		log.warning("read_type_3: about: no file extension")
		return
	return {
		"about_extension": aboutExt,
		"about": aboutContents,
	}


def utf16InfoDecode(b_value):
	"""
		b_value is byte array
		returns str, or None (on errors)

		block type = 3
		block format: <2 byte code1><2 byte code2>
		if code2 == 0: then the block ends
		if code2 == 1: then the block continues as follows:
		<4 byte len1> \x00 \x00 <message in utf-16>
		len1 - length of message in 2-byte chars
	"""
	if b_value[0] != 0:
		log.warning(
			"utf16InfoDecode: b_value=%s, null expected at 0" % list(b_value)
		)
		return

	if b_value[1] == 0:
		if len(b_value) > 2:
			log.warning(
				"utf16InfoDecode: unexpected b_value size: %s" % len(b_value)
			)
		return

	elif b_value[1] > 1:
		log.warning(
			"utf16InfoDecode: b_value=%s, unexpected byte at 1" % list(b_value)
		)
		return

	# now b_value[1] == 1
	size = 2 * binStrToInt(b_value[2:6])
	if tuple(b_value[6:8]) != (0, 0):
		log.warning(
			"utf16InfoDecode: b_value=%s, null expected at 6:8" % list(b_value)
		)
	if size != len(b_value)-8:
		log.warning(
			"utf16InfoDecode: b_value=%s, size does not match" % list(b_value)
		)

	return b_value[8:].decode("utf16")  # str


def flagsInfoDecode(b_value):
	"""
		returns a dict with these keys:
			utf8Encoding
				when this flag is set utf8 encoding is used for all articles
				when false, the encoding is set according to the source and
					target alphabet
			spellingAlternatives
				determines whether the glossary offers spelling alternatives
				for searched terms
			caseSensitive
				defines if the search for terms in this glossary is
					case sensitive
				see code 0x20 as well

	"""
	flags = binStrToInt(b_value)
	return {
		"utf8Encoding": (flags & 0x8000 != 0),
		"spellingAlternatives": (flags & 0x10000 == 0),
		"caseSensitive": (flags & 0x1000 != 0),
	}


infoKeysByCode = {
	0x01: "title",  # glossary name
	0x02: "author",  # glossary author name, a list of "|"-separated values
	0x03: "email",  # glossary author e-mail
	0x04: "copyright",  # copyright message
	0x07: "sourceLang",
	0x08: "targetLang",
	0x09: "description",  # Glossary description
	0x0a: "browsingEnabled",  # 0: browsing disabled, 1: browsing enabled
	0x0b: "iconData",  # FIXME
	0x0c: "bgl_numEntries",
	0x11: "flags",  # the value is a dict
	0x14: "creationTime",
	0x1a: "sourceCharset",
	0x1b: "targetCharset",
	0x1c: "middleUpdated",
	0x2c: "purchaseLicenseMsg",
	0x2d: "licenseExpiredMsg",
	0x2e: "purchaseAddress",
	0x30: "titleWide",
	0x31: "authorWide",
	0x33: "lastUpdated",
	0x3b: "contractions",
	0x3d: "fontName",  # contains a value like "Arial Unicode MS" or "Tahoma"
	0x41: "about",  # (aboutExtention, aboutContents)
	0x43: "length",  # the length of the substring match in a term
}


infoKeyDecodeMethods = {
	"sourceLang": languageInfoDecode,
	"targetLang": languageInfoDecode,
	"browsingEnabled": lambda b_value: (b_value[0] != 0),
	"bgl_numEntries": binStrToInt,
	"creationTime": decodeBglBinTime,
	"middleUpdated": decodeBglBinTime,
	"lastUpdated": decodeBglBinTime,
	"sourceCharset": charsetInfoDecode,
	"targetCharset": charsetInfoDecode,
	"about": aboutInfoDecode,
	"length": binStrToInt,
	"purchaseLicenseMsg": utf16InfoDecode,
	"licenseExpiredMsg": utf16InfoDecode,
	"licenseExpiredMsg": utf16InfoDecode,
	"titleWide": utf16InfoDecode,
	"authorWide": utf16InfoDecode,  # a list of "|"-separated values
	"flags": flagsInfoDecode,
}

"""
bgl_numEntries (0x0c):
	bgl_numEntries does not always matches the number of entries in the
		dictionary, but it's close to it.
	the difference is usually +- 1 or 2, in rare cases may be 9, 29 and more
length (0x43)
	The length of the substring match in a term.
	For example, if your glossary contains the term "Dog" and the substring
		length is 2,
	search of the substrings "Do" or "og" will retrieve the term dog.
	Use substring length 0 for exact match.
contractions (0x3b):
	contains a value like this:
	V-0#Verb|V-0.0#|V-0.1#Infinitive|V-0.1.1#|V-1.0#|V-1.1#|V-1.1.1#Present Simple|V-1.1.2#Present Simple (3rd pers. sing.)|V-2.0#|V-2.1#|V-2.1.1#Past Simple|V-3.0#|V-3.1#|V-3.1.1#Present Participle|V-4.0#|V-4.1#|V-4.1.1#Past Participle|V-5.0#|V-5.1#|V-5.1.1#Future|V2-0#|V2-0.0#|V2-0.1#Infinitive|V2-0.1.1#|V2-1.0#|V2-1.1#|V2-1.1.1#Present Simple (1st pers. sing.)|V2-1.1.2#Present Simple (2nd pers. sing. & plural forms)|V2-1.1.3#Present Simple (3rd pers. sing.)|V2-2.0#|V2-2.1#|V2-2.1.1#Past Simple (1st & 3rd pers. sing.)|V2-2.1.2#Past Simple (2nd pers. sing. & plural forms)|V2-3.0#|V2-3.1#|V2-3.1.1#Present Participle|V2-4.0#|V2-4.1#|V2-4.1.1#Past Participle|V2-5.0#|V2-5.1#|V2-5.1.1#Future||N-0#Noun|N-1.0#|N-1.1#|N-1.1.1#Singular|N-2.0#|N-2.1#|N-2.1.1#Plural|N4-1.0#|N4-1.1#|N4-1.1.1#Singular Masc.|N4-1.1.2#Singular Fem.|N4-2.0#|N4-2.1#|N4-2.1.1#Plural Masc.|N4-2.1.2#Plural Fem.||ADJ-0#Adjective|ADJ-1.0#|ADJ-1.1#|ADJ-1.1.1#Adjective|ADJ-1.1.2#Comparative|ADJ-1.1.3#Superlative||
	value format: (<contraction> "#" [<value>] "|")+
	The value is in second language, that is for Babylon Russian-English.BGL
		the value in russian,
	for Babylon English-Spanish.BGL the value is spanish (I guess), etc.
Glossary manual file (0x41)
	additional information about the dictionary
	in .txt format this may be short info like this:

	Biology Glossary
	Author name: Hafez Divandari
	Author email: hafezdivandari@gmail.com
	-------------------------------------------
	A functional glossary for translating
	English biological articles to fluent Farsi
	-------------------------------------------
	Copyright (c) 2009 All rights reserved.

	in .pdf format this may be a quite large document (about 30 pages),
	an introduction into the dictionary. It describing structure of an article,
	editors, how to use the dictionary.

	format <file extension> "\x00" <file contents>
	file extension may be: ".txt", ".pdf"

purchaseLicenseMsg (0x2c):
	contains a value like this:
	In order to view this glossary, you must purchase a license.
	<br /><a href="http://www.babylon.com/redirects/purchase.cgi?type=170&trid=BPCWHAR">Click here</a> to purchase.

licenseExpiredMsg (0x2d):
	contains a value like this:
	Your license for this glossary has expired.
	In order to view this glossary, you must have a valid license. <br><a href="http://www.babylon.com/redirects/purchase.cgi?type=130&trid=BPCBRTBR">Renew</a> your license today.

purchaseAddress (0x2e):
	contains a value like this:
	http://www.babylon.com/redirects/purchase.cgi?type=169&trid=BPCOT
	or
	mailto:larousse@babylon.com

#elif keyCode==0x20:
#	# 0x30 - case sensitive search is disabled
#	# 0x31 - case sensitive search is enabled
#	# see code 0x11 as well
#	if b_value:
#		value = b_value[0]
"""



# -*- coding: utf-8 -*-
#
# Copyright © 2008-2016 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Copyright © 2011-2012 kubtek <kubtek@gmail.com>
# This file is part of PyGlossary project, http://github.com/ilius/pyglossary
# Thanks to Raul Fernandes <rgfbr@yahoo.com.br> and Karl Grill
#	   for reverse engineering
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

from .bgl_reader import BglReader


class MetaData(object):
	def __init__(self):
		self.blocks = []
		self.numEntries = None
		self.numBlocks = None
		self.numFiles = None
		self.gzipStartOffset = None
		self.gzipEndOffset = None
		self.fileSize = None
		self.bglHeader = None  # data before gzip header


class MetaDataBlock(object):
	def __init__(self, data, _type):
		self.data = data
		self.type = _type


class MetaDataRange(object):
	def __init__(self, _type, count):
		self.type = _type
		self.count = count


class MetaData2(object):
	"""
	Second pass metadata.
	We need to scan all definitions in order to collect these
	statistical data.
	"""
	def __init__(self):
		# defiTrailingFields[i] - number of fields with code i found
		self.defiTrailingFields = [0] * 256
		self.isDefiASCII = True
		# isDefiASCII = true if all definitions contain only ASCII chars
		"""
		We apply a number of tests to each definition, excluding those with
		overwritten encoding (they start with <charset c=U>).
		defiProcessedCount - total number of definitions processed
		defiUtf8Count - number of definitions in utf8 encoding
		defiAsciiCount - number of definitions containing only ASCII chars
		"""
		self.defiProcessedCount = 0
		self.defiUtf8Count = 0
		self.defiAsciiCount = 0
		self.charRefs = dict()  # encoding -> [ 0 ] * 257


class GzipWithCheck(object):
	"""
	gzip.GzipFile with check.
	It checks that unpacked data match what was packed.
	"""
	def __init__(self, fileobj, unpackedPath, reader, closeFileobj=False):
		"""
		constructor

		fileobj - gzip file - archive
		unpackedPath - path of a file containing original data, for testing.
		reader - reference to BglReader class instance, used for logging.
		"""
		self.file = BGLGzipFile(
			fileobj=fileobj,
			closeFileobj=closeFileobj,
		)
		self.unpackedFile = open(unpackedPath, "rb")
		self.reader = reader

	def __del__(self):
		self.close()

	def close(self):
		if self.file:
			self.file.close()
			self.file = None
		if self.unpackedFile:
			self.unpackedFile.close()
			self.unpackedFile = None

	def read(self, size=-1):
		buf1 = self.file.read(size)
		buf2 = self.unpackedFile.read(size)
		if buf1 != buf2:
			self.reader.msgLogFileWrite(
				"GzipWithCheck.read: !=: size = %s, (%s) (%s)" % (
					buf1,
					buf2,
					size,
				)
			)
#		else:
#			self.reader.msgLogFileWrite(
#				"GzipWithCheck.read: ==: size = %s, (%s) (%s)" % (
#					buf1,
#					buf2,
#					size,
#				)
#			)
		return buf1

	def seek(self, offset, whence=os.SEEK_SET):
		self.file.seek(offset, whence)
		self.unpackedFile.seek(offset, whence)
#		self.reader.msgLogFileWrite(
#			"GzipWithCheck.seek: offset = %s, whence = %s" % (offset, whence)
#		)

	def tell(self):
		pos1 = self.file.tell()
		pos2 = self.unpackedFile.tell()
		if pos1 != pos2:
			self.reader.msgLogFileWrite(
				"GzipWithCheck.tell: !=: %s %s" % (pos1, pos2)
			)
#		else:
#			self.reader.msgLogFileWrite(
#				"GzipWithCheck.tell: ==: %s %s" % (pos1, pos2)
#			)
		return pos1

	def flush(self):
		if os.sep == "\\":
			pass
			# a bug in Windows
			# after file.flush, file.read returns garbage
		else:
			self.file.flush()
			self.unpackedFile.flush()


class DebugBglReader(BglReader):
	def open(
		self,
		filename,
		collectMetadata2=False,
		searchCharSamples=False,
		writeGz=False,
		rawDumpPath=None,
		unpackedGzipPath=None,
		charSamplesPath=None,
		msgLogPath=None,
		**kwargs
	):
		if not BglReader.open(self, filename, **kwargs):
			return

		self.metadata2 = MetaData2() if collectMetadata2 else None
		self.targetCharsArray = ([False] * 256) if searchCharSamples else None

		self.writeGz = writeGz
		self.rawDumpPath = rawDumpPath
		self.unpackedGzipPath = unpackedGzipPath
		self.charSamplesPath = charSamplesPath
		self.msgLogPath = msgLogPath

		if self.rawDumpPath:
			self.rawDumpFile = open(self.rawDumpPath, "w")
		if self.charSamplesPath:
			self.samplesDumpFile = open(self.charSamplesPath, "w")
		if self.msgLogPath:
			self.msgLogFile = open(self.msgLogPath, "w")

		self.charRefStatPattern = re.compile(b"(&#\\w+;)", re.I)

	def openGzip(self):
		with open(self._filename, "rb") as bglFile:
			if not bglFile:
				log.error("file pointer empty: %s" % bglFile)
				return False
			buf = bglFile.read(6)
			if len(buf) < 6 or not buf[:4] in (
				b"\x12\x34\x00\x01",
				b"\x12\x34\x00\x02",
			):
				log.error("invalid header: %s" % buf[:6])
				return False
			self.gzipOffset = gzipOffset = binStrToInt(buf[4:6])
			log.debug("Position of gz header: i=%s" % gzipOffset)
			if gzipOffset < 6:
				log.error("invalid gzip header position: %s" % gzipOffset)
				return False

			if self.writeGz:
				self.dataFile = self._filename+"-data.gz"
				try:
					f2 = open(self.dataFile, "wb")
				except IOError:
					log.exception("error while opening gzip data file")
					self.dataFile = join(
						tmpDir,
						os.path.split(self.m_filename)[-1] + "-data.gz"
					)
					f2 = open(self.dataFile, "wb")
				bglFile.seek(i)
				f2.write(bglFile.read())
				f2.close()
				self.file = gzip.open(self.dataFile, "rb")
			else:
				f2 = FileOffS(self._filename, gzipOffset)
				if self.unpackedGzipPath:
					self.file = GzipWithCheck(
						f2,
						self.unpackedGzipPath,
						self,
						closeFileobj=True,
					)
				else:
					self.file = BGLGzipFile(
						fileobj=f2,
						closeFileobj=True,
					)

	def close(self):
		BglReader.close(self)
		if self.rawDumpFile:
			self.rawDumpFile.close()
			self.rawDumpFile = None
		if self.msgLogFile:
			self.msgLogFile.close()
			self.msgLogFile = None
		if self.samplesDumpFile:
			self.samplesDumpFile.close()
			self.samplesDumpFile = None

	def __del__(self):
		BglReader.__del__(self)

	def readEntryWord(self, block, pos):
		succeed, pos, u_word, b_word = \
			BglReader.readEntryWord(self, block, pos)
		if not succeed:
			return
		self.rawDumpFileWriteText("\n\nblock type = %s\nkey = " % block.type)
		self.rawDumpFileWriteData(b_word)

	def readEntryDefi(self, block, pos, b_key):
		succeed, pos, u_defi, b_defi = \
			BglReader.readEntryDefi(self, block, pos, b_key)
		if not succeed:
			return
		self.rawDumpFileWriteText("\ndefi = ")
		self.rawDumpFileWriteData(b_defi)

	"""
	def readEntryAlts(self, block, pos, b_key, key):
		succeed, pos, alts, b_alts = \
			BglReader.readEntryAlts(self, block, pos, b_key, key)
		if not succeed:
			return
		for b_alt in b_alts:
			self.rawDumpFileWriteText("\nalt = ")
			self.rawDumpFileWriteData(b_alt)
	"""

	def charReferencesStat(self, b_text, encoding):
		"""
			b_text is bytes instance
		"""
		# &#0147;
		# &#x010b;
		if not self.metadata2:
			return

		if encoding not in self.metadata2.charRefs:
			self.metadata2.charRefs[encoding] = [0] * 257
		charRefs = self.metadata2.charRefs[encoding]

		for index, b_part in enumerate(re.split(
			self.charRefStatPattern,
			b_text,
		)):
			if index % 2 != 1:
				continue
			try:
				if b_part[:3].lower() == "&#x":
					code = int(b_part[3:-1], 16)
				else:
					code = int(b_part[2:-1])
			except (ValueError, OverflowError):
				continue
			if code <= 0:
				continue
			code = min(code, 256)
			charRefs[code] += 1

	def processDefiStat(self, fields, b_defi, b_key):
		if fields.singleEncoding:
			self.findAndPrintCharSamples(
				fields.b_defi,
				"defi, key = %s" + b_key,
				fields.encoding,
			)
			if self.metadata2:
				self.metadata2.defiProcessedCount += 1
				if isASCII(fields.b_defi):
					self.metadata2.defiAsciiCount += 1
				try:
					fields.b_defi.decode("utf8")
				except UnicodeError:
					pass
				else:
					self.metadata2.defiUtf8Count += 1
		if self.metadata2 and self.metadata2.isDefiASCII:
			if not isASCII(fields.u_defi):
				self.metadata2.isDefiASCII = False

	# write text to dump file as is
	def rawDumpFileWriteText(self, text):  # FIXME
		text = toStr(text)
		if self.rawDumpFile:
			self.rawDumpFile.write(text)

	# write data to dump file unambiguously representing control chars
	# escape "\" with "\\"
	# print control chars as "\xhh"
	def rawDumpFileWriteData(self, text):
		text = toStr(text)
		# the next function escapes too many chars, for example, it escapes äöü
		# self.rawDumpFile.write(text.encode("unicode_escape"))
		if self.rawDumpFile:
			self.rawDumpFile.write(text)

	def msgLogFileWrite(self, text):
		text = toStr(text)
		if self.msgLogFile:
			offset = self.msgLogFile.tell()
			# print offset in the log file to facilitate navigating this
			# log in hex editor
			# intended usage:
			# the log file is opened in a text editor and hex editor
			# use text editor to read error messages, use hex editor to
			# inspect char codes offsets allows to quickly jump to the right
			# place of the file hex editor
			self.msgLogFile.write("\noffset = {0:#X}\n" % offset)
			self.msgLogFile.write(text+"\n")
		else:
			log.debug(text)

	def samplesDumpFileWrite(self, text):
		text = toStr(text)
		if self.samplesDumpFile:
			offset = self.samplesDumpFile.tell()
			self.samplesDumpFile.write("\noffset = {0:#X}\n" % offset)
			self.samplesDumpFile.write(text+"\n")
		else:
			log.debug(text)

	def dumpBlocks(self, dumpPath):
		import pickle
		self.file.seek(0)
		metaData = MetaData()
		metaData.numFiles = 0
		metaData.gzipStartOffset = self.gzipOffset

		self.numEntries = 0
		self.numBlocks = 0
		range_type = None
		range_count = 0
		block = Block()
		while not self.isEndOfDictData():
			log.debug(
				"readBlock: " +
				"offset %#X, " % self.file.tell() +
				"unpacked offset %#X" % self.file.unpackedFile.tell()
			)
			if not self.readBlock(block):
				break
			self.numBlocks += 1

			if block.type in (1, 7, 10, 11, 13):
				self.numEntries += 1
			elif block.type == 2:  # Embedded File (mostly Image or HTML)
				metaData.numFiles += 1

			if block.type in (1, 2, 7, 10, 11, 13):
				if range_type == block.type:
					range_count += 1
				else:
					if range_count > 0:
						mblock = MetaDataRange(range_type, range_count)
						metaData.blocks.append(mblock)
						range_count = 0
					range_type = block.type
					range_count = 1
			else:
				if range_count > 0:
					mblock = MetaDataRange(range_type, range_count)
					metaData.blocks.append(mblock)
					range_count = 0
				mblock = MetaDataBlock(block.data, block.type)
				metaData.blocks.append(mblock)

		if range_count > 0:
			mblock = MetaDataRange(range_type, range_count)
			metaData.blocks.append(mblock)
			range_count = 0

		metaData.numEntries = self.numEntries
		metaData.numBlocks = self.numBlocks
		metaData.gzipEndOffset = self.file_bgl.tell()
		metaData.fileSize = os.path.getsize(self._filename)
		with open(self._filename, "rb") as f:
			metaData.bglHeader = f.read(self.gzipOffset)

		with open(dumpPath, "wb") as f:
			pickle.dump(metaData, f)

		self.file.seek(0)

	def dumpMetadata2(self, dumpPath):
		import pickle
		if not self.metadata2:
			return
		with open(dumpPath, "wb") as f:
			pickle.dump(self.metadata2, f)

	def processDefiStat(self, fields, defi, b_key):
		BglReader.processDefiStat(self, fields, defi, b_key)

		if fields.b_title:
			self.rawDumpFileWriteText("\ndefi title: ")
			self.rawDumpFileWriteData(fields.b_title)
		if fields.b_title_trans:
			self.rawDumpFileWriteText("\ndefi title trans: ")
			self.rawDumpFileWriteData(fields.b_title_trans)
		if fields.b_transcription_50:
			self.rawDumpFileWriteText(
				"\ndefi transcription_50 (%#x): "
				% fields.code_transcription_50
			)
			self.rawDumpFileWriteData(fields.b_transcription_50)
		if fields.b_transcription_60:
			self.rawDumpFileWriteText(
				"\ndefi transcription_60 (%#x): "
				% fields.code_transcription_60
			)
			self.rawDumpFileWriteData(fields.b_transcription_60)
		if fields.b_field_1a:
			self.rawDumpFileWriteText("\ndefi field_1a: ")
			self.rawDumpFileWriteData(fields.b_field_1a)
		if fields.b_field_13:
			self.rawDumpFileWriteText(
				"\ndefi field_13 bytes: %r" % fields.b_field_13
			)
		if fields.b_field_07:
			self.rawDumpFileWriteText("\ndefi field_07: ")
			self.rawDumpFileWriteData(fields.b_field_07)
		if fields.b_field_06:
			self.rawDumpFileWriteText(
				"\ndefi field_06: %s" % fields.b_field_06
			)

	# search for new chars in data
	# if new chars are found, mark them with a special sequence in the text
	# and print result into msg log
	def findAndPrintCharSamples(self, b_data, hint, encoding):
		assert isinstance(b_data, bytes)
		if not self.targetCharsArray:
			return
		offsets = self.findCharSamples(b_data)
		if len(offsets) == 0:
			return
		res = ""
		utf8 = (encoding.lower() == "utf8")
		i = 0
		for o in offsets:
			j = o
			if utf8:
				while b_data[j] & 0xc0 == 0x80:
					j -= 1
			res += b_data[i:j]
			res += "!!!--+!!!"
			i = j
		res += b_data[j:]
		offsets_str = " ".join([str(el) for el in offsets])
		self.samplesDumpFileWrite(
			"charSample(%s)\noffsets = %s\nmarked = %s\norig = %s\n" % (
				hint,
				offsets_str,
				res,
				b_data,
			)
		)

	def findCharSamples(self, b_data):
		"""
		Find samples of chars in b_data.

		Search for chars in data that have not been marked so far in
		the targetCharsArray array, mark new chars.
		Returns a list of offsets in b_data
		May return an empty list.
		"""
		res = []
		if not isinstance(b_data, bytes):
			log.error("findCharSamples: b_data is not a bytes instance")
			return res
		if not self.targetCharsArray:
			log.error("findCharSamples: self.targetCharsArray == None")
			return res
		for i, char in enumerate(b_data):
			if x < 128:
				continue
			if not self.targetCharsArray[x]:
				self.targetCharsArray[x] = True
				res.append(i)
		return res



# -*- coding: utf-8 -*-
# octopus_mdic.py
# Read Octopus MDict dictionary format, mdx(dictionary)/mdd(data)
#
# Copyright (C) 2013 Xiaoqiang Wang <xiaoqiangwang AT gmail DOT com>
# Copyright (C) 2013-2016 Saeed Rasooli <saeed.gnu@gmail.com>
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

from formats_common import *

import os
from os.path import splitext, isfile, isdir, extsep, basename, dirname

enable = True
format = "OctopusMdict"
description = "Octopus MDict"
extentions = [".mdx"]
readOptions = [
	"encoding",  # str
	"substyle",  # bool
]
writeOptions = []


class Reader(object):
	def __init__(self, glos):
		self._glos = glos
		self.clear()

	def clear(self):
		self._filename = ""
		self._encoding = ""
		self._substyle = True
		self._mdx = None
		self._mdd = None
		self._mddFilename = ""

	def open(self, filename, **options):
		from pyglossary.plugin_lib.readmdict import MDX, MDD
		self._filename = filename
		self._encoding = options.get("encoding", "")
		self._substyle = options.get("substyle", True)
		self._mdx = MDX(filename, self._encoding, self._substyle)

		filenameNoExt, ext = splitext(self._filename)
		mddFilename = "".join([filenameNoExt, extsep, "mdd"])
		if isfile(mddFilename):
			self._mdd = MDD(mddFilename)
			self._mddFilename = mddFilename

		log.debug("mdx.header = " + pformat(self._mdx.header))
		# for key, value in self._mdx.header.items():
		#	key = key.lower()
		#	self._glos.setInfo(key, value)
		try:
			title = self._mdx.header[b"Title"]
		except KeyError:
			pass
		else:
			self._glos.setInfo("title", title)
		self._glos.setInfo(
			"description",
			self._mdx.header.get(b"Description", ""),
		)

	def __iter__(self):
		if self._mdx is None:
			log.error("trying to iterate on a closed MDX file")
		else:
			for word, defi in self._mdx.items():
				word = toStr(word)
				defi = toStr(defi)
				yield self._glos.newEntry(word, defi)
			self._mdx = None

		if self._mdd:
			for b_fname, b_data in self._mdd.items():
				fname = toStr(b_fname)
				fname = fname.replace("\\", os.sep).lstrip(os.sep)
				yield self._glos.newDataEntry(fname, b_data)
			self._mdd = None

	def __len__(self):
		if self._mdx is None:
			log.error(
				"OctopusMdict: called len(reader) while reader is not open"
			)
			return 0
		return len(self._mdx)

	def close(self):
		self.clear()



# -*- coding: utf-8 -*-

from formats_common import *

enable = True
format = 'Sql'
description = 'SQL'
extentions = ['.sql']
readOptions = []
writeOptions = [
	'encoding',  # str
]


def write(
	glos,
	filename,
	encoding='utf-8',
):
	with open(filename, 'w', encoding=encoding) as fp:
		for line in glos.iterSqlLines(
			transaction=False,
		):
			fp.write(line + '\n')



import re
import string

parenthetical = re.compile(r"\([^)]+?\)")
punct_table = {ord(p): " " for p in string.punctuation if p not in "-'"}
stops = {"i","me","my","myself","we","our","ours","ourselves","you","your","yours","yourself","yourselves","he","him","his","himself","she","her","hers","herself","it","its","itself","they","them","their","theirs","themselves","what","which","who","whom","this","that","these","those","am","is","are","was","were","be","been","being","have","has","had","having","do","does","did","doing","a","an","the","and","but","if","or","because","as","until","while","of","at","by","for","with","about","against","between","into","through","during","before","after","above","below","to","from","up","down","in","out","on","off","over","under","again","further","then","once","here","there","when","where","why","how","all","any","both","each","few","more","most","other","some","such","no","nor","not","only","own","same","so","than","too","very","s","t","can","will","just","don","should","now","d","ll","m","o","re","ve","y","ain","aren","couldn","didn","doesn","hadn","hasn","haven","isn","ma","mightn","mustn","needn","shan","shouldn","wasn","weren","won","wouldn"}
def summarize(phrase):
	phrase = parenthetical.sub("", phrase)
	phrase = phrase.translate(punct_table)
	words = phrase.split()
	relevant_words = [word for word in words if word not in stops]
	if not relevant_words:
		relevant_words = words
	summary = " ".join(relevant_words[:10])
	return summary




def cleanWinArabicStr(u):
	"""
		u is a utf-8 encoded string
	"""
	replaceList = [
		('ی', 'ي'),
		('ک', 'ك'),
		('ٔ', 'ء'),
		('\ufffd', ''),

	] + [(chr(i), chr(i+144)) for i in range(1632, 1642)]
	for item in replaceList:
		u = u.replace(item[0], item[1])
	return u


def recodeToWinArabic(u):
	"""
		u is a utf-8 encoded string
	"""
	u = cleanWinArabicStr(u)
	return u.encode('windows-1256', 'replace')
